/*****************************************************************************/
/*
                                Cache.c

This module implements a file data and revision time cache.  It is different to
most other modules in that it doesn't have a "task" structure.  The small
amount of storage required is integrated into the request structure. It is
designed to provided an efficient static document request (file transfer)
mechanism.  Unlike the file module which may interleave it's activities within
the those of other modules (e.g. the directory module using it to provide
read-me information), it can only be used once, stand-alone per request.

Cache data is loaded by the file module while concurrently transfering that
data to the request client, using buffer space supplied by cache module, space
that can then be retained for reuse as cache. Hence the cache load adds no
significant overhead to the actual reading and initial transfer of the file.

The cache has limits imposed on it.  A maximum number of files that can be
cached at any one time, and a maximum amount of memory that can be allocated
in total by the cache.  Each of these two operates to limit the other.  In
addition the maximum size of a file that can be loaded into the cache can be
specified.

Terminology
-----------

"hit" refers to a request path being found in cache. If the data is still
valid the request can be supplied from cache.

"load"ing the cache refers to reading the contents of a file into cache
memory.

"valid" means that the file from which the cached data was originally read has
not had it's revision date changed (the implication being is the file contents
have not changed.

Why Implement Caching?
----------------------

Caching, be definition, attempts to improve performance by keeping data in
storage that is faster to access than it's usual location.  The performance
improvement can be assessed in three basic ways.  Reduction in latency when
accessing the data, of processing involved, and in impact on the usual storage
location.

This cache is provided to address all three.  Where networks are particularly
responsive a reduction in request latency can often be noticable.  Where
servers are particularly busy or where disk systems particularly loaded a
reduction in the need to access the file system can significantly improve
performance.  My suggestion is though, that for most VMS sites high levels of
hits are not a great concern, and for these caching can easily be left
disabled.
            
Why take so long to implement caching? Well, WASD's intranet services are not
particularly busy sites. This coupled with powerful hardware has meant
performance has not been an overriding concern. However, this cache module
came about because I felt like creating it and it was an obvious lack of
functionality within the server, not because WASD (the organisation) needed
it.

Cache Suitability Considerations
--------------------------------

A cache is not always of benefit!  It's cost may outweigh it's return.

Any cache's efficiencies can only occur where subsets of data are consistently
being demanded. Although these subsets may change slowly over time a consistent
and rapidly changing aggregate of requests lose the benefit of more readily
accessable data to the overhead of cache management due to the constant and
continuous flushing and reloading of cache data. This server's cache is no
different, it will only improve performance if the site experiences some
consistency in the files requested. For sites that have only a small
percentage of files being repeatedly requested it is probably better that the
cache be disabled. The other major consideration is available system memory.
On a system where memory demand is high there is little value in having cache
memory sitting in page space, trading disk I/O and latency for paging I/O and
latency. On memory-challenged systems cache is probably best disabled.

To help assessment of the cache's efficiency for any given site monitor the
administration menu's cache report.

With "loads not hit", the count represents the cumulative number of files
loaded but never subsequently hit. If this percentage is high it means most
files loaded are never hit, indicating the site's request profile is possibly
unsuitable for caching.

The item "hits" respresents the cumulative, total number of hits against the
cumulative, total number of loads. The percentage here can range from zero to
many thousands of percent :^) with less than 100% indicating poor cache
performance, from 200% upwards better and good performance. The items "1-9",
"10-99" and "100+" show the count and percentage of total hits that occured
when a given entry had experienced hits within that range (e.g. if an entry
has had 8 previous hits, the ninth increments the "1-9" item whereas the tenth
and eleventh increments the "10-99" item, etc.)

Other considerations also apply when assessing the benefit of having a cache.
For example, a high number and percentage of hits can be generated while the
percentage of "loads not hit" could be in the also be very high. The
explanation for this would be one or two frequently requested files being hit
while most others are loaded, never hit, and flushed as other files request
cache space. In situations such as this it is difficult to judge whether cache
processing is improving performance or just adding overhead.

Again, my suggestion is, that for most VMS sites, high levels of access are not
a great concern, and for these caching can easily be left disabled.
            
Description
-----------

An MD5 digest (16 byte hash) is used to uniquely identify cached files.  The
hash is generated either from a mapped path or from the file name before
calling any of the cache search, load or completion functions.  This MD5 hash
is then stored along with the file details and contents so that identical
paths/files can be matched during subsequent searches.  The MD5 algorithm
"guarantees" a unique identifier for any resource name, and the 16 byte size
makes matching using just 4 longword comparisons very efficient.

If using a path it MUST be the mapped path, not the client-supplied, request
path.  With conditional mapping, identical request paths may be mapped to
completely different virtual paths.

Space for a file's data is dynamically allocated and reallocated if necessary
as cache entries are reused.  It is allocated in user-specifiable chunks.  It
is expected this mechanism provides some efficiencies when reusing cache
entries.

A simple hash table is used to try and initially hit an entry.  A collision
list allows rapid subsequent searching.  The hash table is a fixed 4096 entries
with the hash value generated by directly using a fixed three bytes of the
MD5 hash for a range from 0..4095.

Cache entries are also maintained in a global linked list with the most recent
and most frequently hit entries towards the head of the list. The linked-list
organisation allows a simple implementation of a least-recently-used (LRU)
algorithm for selecting an entry when a new request demands an entry and space
for cache loading. The linked list is naturally ordered from most recently and
most frequently accessed at the head, to the least recently and least
frequently accessed at the tail. Hence an infrequently accessed entry is
selected from the tail end of the list, it's data invalidated and given to the
new request for cache load. Invalidated data cache entries are also
immediately placed at the tail of the list for reuse/reloading.

When a new entry is initially loaded it is placed at the top of the list. Hits
on other entries result in a check being made against the number of hits of
head entry in the list. If the entry being hit has a higher hit count it is
placed at the head of the list, pushing the previously head entry "down". If
not then it is again checked against the entry immediately before it in the
list. If higher then the two are swapped. This results in the most recently
loaded entries and the more frequently hit being nearest and migrating towards
the start of the search.

To help prevent the cache thrashing with floods of requests for not currently
loaded files, any entry that has a suitably high number of hits over the recent
past (suitably high ... how many is that, and recent past ... how long is
that?) are not reused until no hits have occured within that period.  Hopefully
this prevents lots of unnecessary loads of one-offs at the expense of genuinely
frequently accessed files.

To prevent multiple loads of the same path/file, for instance if a subsequent
request demands the same file as a previous request is still currently loading,
any subsequent request will merely transfer the file, not concurrently load it
into the cache.

Cache Content Validation
------------------------

The cache will automatically revalidate the file data after a specified number
of seconds by comparing the original file revision time to the current revision
time. If different the file contents have changed and the cache contents
declared invalid. If found invalid the file transfer then continues outside of
the cache with the new contents being concurrently reloaded into the cache.
Cache validation is also always performed if the request uses "Pragma:
no-cache" (i.e. as with the Netscape Navigator reload function). Hence there is
no need for any explicit flushing of the cache under normal operation. If a
document does not immediately reflect and changes made to it (i.e. validation
time has not been reached) validation (and consequent reload) can be "forced"
with a browser reload. The entire cache may be purged of cached data either
from the server administration menu or using command line server control.

Permanent Entries
-----------------

Permanent entries are indicated by a path mapping a SET rule.  Permanent
entries are designed to allow certain classes of file, those that are
relatively static and/or frequently being used by requests, that once loaded
remain in the file cache, never validated, never able to be scavenged or
otherwise removed from the cache during routine server activity.  These
entries, along with the volatile ones, can of course be purged from the cache
either via the CLI /DO=CACHE=PURGE command or using Server Administration menu. 
Permanent entries do not use any of the data memory ([CacheTotalKBytesMax]) set
aside for the volatile cache.  They use memory from their own specific VM pool. 
Unlike for volatile entries there is no upper limit (apart from system and
server process virtual memory) on the memory that can be allocated for
permanent entries (be careful!).  Entry slots ([CacheEntriesMax]) are shared
between permanent and volatile entries (and can be a constraint on both).


VERSION HISTORY
---------------
16-JUN-2003  MGD  bugfix; FileSetCharset() moved from FILE.C module
24-MAY-2003  MGD  permanent cache entries,
                  path specified maximum file size
11-MAR-2002  MGD  bugfix; ensure only one request revalidates a cache entry at
                  a time (multiple could cause eventual channel exhaustion)
22-NOV-2001  MGD  ensure there are reasonable cache minima
29-SEP-2001  MGD  instance support
04-AUG-2001  MGD  use MD5 hash to identify cache entries,
                  modify hash table as fixed 4096 entrie,
                  support module WATCHing
27-MAY-2000  MGD  BUGFIX; CacheEntryNotValid() linked list :^{
                  bugfix; CacheLoadBegin() #else before memcpy()
09-APR-2000  MGD  simplified cache search
04-MAR-2000  MGD  use NetWriteFaol(), et.al.
28-DEC-1999  MGD  support ODS-2 and ODS-5 using ODS module,
                  add a number of other WATCH points
26-SEP-1999  MGD  minor changes in line with RequestExecute(),
                  CacheReport() now only optionally reports cached files,
                  scavenge failure should not result in a sanity check exit
20-JAN-1999  MGD  report format refinenment
19-SEP-1998  MGD  improve granularity of cache operation,
                  add check for existing entry to CacheLoadBegin()
14-MAY-1998  MGD  request-specified content-type ("httpd=content&type=")
18-MAR-1998  MGD  use file's VBN and first free byte to check size changes
                  (allows variable record files to be cached more efficiently)
24-FEB-1998  MGD  if CacheAcpInfo() reports a problem
                  then let the file module handle/report it,
                  add file size check to entry validation (allow for extend)
10-JAN-1998  MGD  added a (much overdue) hash collision list,
                  fixed problem with cache purge (it mostly didn't :^)
22-NOV-1997  MGD  sigh, bugfix; need to proactively free memory at capacity
05-OCT-1997  MGD  initial development for v4.5
*/
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <atrdef.h>
#include <descrip.h>
#include <fibdef.h>
#include <iodef.h>
#include <libdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "md5.h"

#define WASD_MODULE "CACHE"

/******************/
/* global storage */
/******************/

BOOL  CacheEnabled,
      CacheHashTableInitialised;

/* this is run-time storage */
int  CacheChunkInBytes,
     CacheEntryCount,
     CacheEntriesMax,
     CacheFileKBytesMax,
     CacheFrequentHits,
     CacheFrequentSeconds,
     CacheHashTableMask,
     CacheMemoryInUse,
     CachePermEntryCount,
     CachePermMemoryInUse,
     CacheTotalKBytesMax,
     CacheValidateSeconds;

/* these counters may be zeroed when accounting is zeroed */
int  CacheHashTableCollsnCount,
     CacheHashTableCount,
     CacheHashTableHitCount,
     CacheHashTableMissCount,
     CacheHashTableCollsnHitCount,
     CacheHashTableCollsnMissCount,
     CacheHitCount,
     CacheHits0,
     CacheHits10,
     CacheHits100,
     CacheHits100plus,
     CacheListHitCount,
     CacheLoadCount,
     CacheMemoryAllocateCount,
     CacheMemoryReuseCount,
     CacheNotHitCount,
     CacheNoHitsCount,
     CacheScavengeCount,
     CacheScavengeFailCount,
     CacheSearchCount;

LIST_HEAD  CacheList;

/* the hash table index is the first byte of the MD5 hash */
#define CACHE_HASH_TABLE_ENTRIES 4096 /* 12 bits, 16384 bytes, 32 page(lets) */
struct GenericCacheEntry  *CacheHashTable [CACHE_HASH_TABLE_ENTRIES];

/********************/
/* external storage */
/********************/

extern BOOL  CliCacheEnabled,
             CliCacheDisabled;

extern int  HttpdTickSecond,
            OdsExtended,
            OutputBufferSize;

extern char  ErrorSanityCheck[],
             HttpProtocol[],
             ServerHostPort[],
             SoftwareID[],
             Utility[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern MSG_STRUCT  Msgs;
extern HTTPD_PROCESS  HttpdProcess;
extern WATCH_STRUCT  Watch;

/*****************************************************************************/
/*
Initialize cache run-time parameters at startup (even though caching might be
initially disabled).  Also allocate the hash table when appropriate.
*/ 

CacheInit (BOOL Startup)

{
   int  HashValue;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "CacheInit()");

   if (Startup)
   {
      if (!CliCacheDisabled && (CliCacheEnabled || Config.cfCache.Enabled))
         CacheEnabled = true;
      else
         CacheEnabled = false;

      CacheChunkInBytes = Config.cfCache.ChunkKBytes << 10;
      if (CacheChunkInBytes < 512) CacheChunkInBytes = 512;
      CacheEntriesMax = Config.cfCache.EntriesMax;
      if (CacheEntriesMax < 256) CacheEntriesMax = 256;
      CacheFileKBytesMax = Config.cfCache.FileKBytesMax;
      if (CacheFileKBytesMax <= 0) CacheFileKBytesMax = 65;
      CacheFrequentHits = Config.cfCache.FrequentHits;
      CacheFrequentSeconds = Config.cfCache.FrequentSeconds;
      CacheTotalKBytesMax = Config.cfCache.TotalKBytesMax;
      if (CacheTotalKBytesMax < 1024) CacheTotalKBytesMax = 1024;
      CacheValidateSeconds = Config.cfCache.ValidateSeconds;

      CacheEntryCount = CacheMemoryInUse =
         CachePermEntryCount = CachePermMemoryInUse = 0;

      CacheZeroCounters ();
   }

   if (CacheEntriesMax && CacheEnabled)
   {
      /* initialize cache virtual memory management */
      VmCacheInit (CacheTotalKBytesMax);
      VmPermCacheInit (CacheTotalKBytesMax);

      for (HashValue = 0; HashValue < CACHE_HASH_TABLE_ENTRIES; HashValue++)
         CacheHashTable[HashValue] = NULL;

      CacheHashTableInitialised = true;
   }
}

/*****************************************************************************/
/*
Just zero the cache-associated counters :^)
*/ 

CacheZeroCounters ()

{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "CacheZeroCounter()");

   CacheHitCount = CacheHashTableHitCount =
      CacheHashTableCollsnCount = CacheHashTableCollsnHitCount =
      CacheHashTableCollsnMissCount = CacheHashTableCount =
      CacheHashTableMissCount =
      CacheMemoryAllocateCount = CacheMemoryReuseCount =
      CacheNoHitsCount = CacheHits0 = CacheHits10 = CacheHits100 =
      CacheHits100plus = CacheLoadCount = 
      CacheScavengeCount = CacheScavengeFailCount = CacheSearchCount = 0;
}

/*****************************************************************************/
/*
Look through the cache list for a resource hash (MD5) that matches the one in
the file task structure.  If one is found then call CacheBegin() and return
success, otherwise return error status to indicate the search was unsuccessful. 
CacheBegin() (actually using CacheFileInfoAst()) may still AST back to
FileEnd() if there is a problem returning the cached contents (i.e. contents
have changed on-disk), it is the resposibility of that routine to continue
processing the request.

NOTE: as with CacheLoadBegin(), the path passed to this function MUST be the
mapped path, not the request path. With conditional mapping identical request
paths may be mapped to completely different virtual paths.
*/ 

int CacheSearch (REQUEST_STRUCT *rqptr)

{
   int  EntryCount,
        HashValue;
   char  *cptr, *sptr;
   FILE_CENTRY  *captr;
   FILE_TASK  *ftkptr;
   LIST_ENTRY  *leptr;
   MD5_HASH  *md5ptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheSearch() !16&H !UL",
                 &rqptr->FileTaskPtr->Md5Hash,
                 rqptr->FileTaskPtr->Md5Hash.HashLong[0] & 0xfff);

   if (!CacheHashTableInitialised) CacheInit (false);

   /* if a SET mapping rule has specified the path should not be cached */
   if (rqptr->rqPathSet.NoCache) return (SS$_NOSUCHFILE);

   /* initialize so there is no request-associated cache entry */
   rqptr->rqCache.EntryPtr = NULL;

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "CACHE search !16&H", &rqptr->FileTaskPtr->Md5Hash);

   /* check the hash table */
   md5ptr = &rqptr->FileTaskPtr->Md5Hash;
   CacheHashTableCount++;
   /* 12 bit hash value, 0..4095 from a fixed 3 bytes of the MD5 hash */
   HashValue = md5ptr->HashLong[0] & 0xfff;
   if (!(captr = CacheHashTable[HashValue]))
   {
      /*************/
      /* not found */
      /*************/

      /* nope, no pointer against that hash value */
      CacheHashTableMissCount++;
      return (SS$_NOSUCHFILE);
   }

   EntryCount = 0;
   for ( /*set above*/ ; captr; captr = captr->HashCollisionNextPtr)
   {
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
"valid:!&B loading:!&B revalidating:!&B !AZ !&?no-match\rmatch\r",
                    captr->CacheValid, captr->CacheLoading,
                    captr->CacheRevalidating, captr->FileOds.ExpFileName,
                    captr->Md5Hash.HashLong[0] != md5ptr->HashLong[0] ||
                    captr->Md5Hash.HashLong[1] != md5ptr->HashLong[1] ||
                    captr->Md5Hash.HashLong[2] != md5ptr->HashLong[2] ||
                    captr->Md5Hash.HashLong[3] != md5ptr->HashLong[3]);

      /* note the first collision entry */
      if (++EntryCount == 2) CacheHashTableCollsnCount++;

      /* match each of the 4 sets of 4 bytes (longwords) in the MD5 hash */
      if (captr->Md5Hash.HashLong[0] != md5ptr->HashLong[0] ||
          captr->Md5Hash.HashLong[1] != md5ptr->HashLong[1] ||
          captr->Md5Hash.HashLong[2] != md5ptr->HashLong[2] ||
          captr->Md5Hash.HashLong[3] != md5ptr->HashLong[3])
         continue; 

      /* if the data is not valid but being loaded then consider it a hit */
      if (!captr->CacheValid && !captr->CacheLoading) continue;

      /********/
      /* hit! */
      /********/

      /* first entry is always the hash table, after that the collision list */
      if (EntryCount == 1)
         CacheHashTableHitCount++;
      else
         CacheHashTableCollsnHitCount++;

      if (!captr->CacheValid ||
          captr->CacheLoading ||
          captr->CacheRevalidating)
      {
         /**************/
         /* not usable */
         /**************/

         rqptr->rqCache.NotUsable = true;
         return (SS$_NOSUCHFILE);
      }

      /**********/
      /* usable */
      /**********/

      rqptr->rqCache.EntryPtr = captr;
      CacheBegin (rqptr);
      return (SS$_NORMAL);
   }

   /***********/
   /* not hit */
   /***********/

   CacheHashTableCollsnMissCount++;
   return (SS$_NOSUCHFILE);
}

/*****************************************************************************/
/*
Use this cached entry as the contents for the client.  If the entry needs to
be validated then generated an asynchronous ACP QIO to get the required file
details, other wise call the AST processing function directly.

NOTE: as with CacheLoadBegin(), the path passed to this function MUST be the
mapped path, not the request path.  With conditional mapping identical request
paths may be mapped to completely different virtual paths.
*/ 
 
CacheBegin (REQUEST_STRUCT *rqptr)

{
   BOOL  RevalidateEntry;
   FILE_CENTRY  *captr;

   /*********/
   /* begin */
   /*********/

   captr = rqptr->rqCache.EntryPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheBegin() !AZ", captr->FileOds.ExpFileName);

   if (captr->CachePermanent)
      RevalidateEntry = false;
   else
   if (rqptr->PragmaNoCache || captr->ValidateTickSecond < HttpdTickSecond)
      RevalidateEntry = true;
   else
      RevalidateEntry = false;

   /* if no need to recheck revision date/time, explicitly call the AST */
   if (!RevalidateEntry)
   {
      /* status of non-zero used to determine whether the ACP QIO was used */
      captr->FileOds.AcpInfo.AcpIOsb.Status = 0;
      CacheAcpInfoAst (rqptr);
      return;
   }

   /* ensure only one request at a time revalidates a cache entry */
   captr->CacheRevalidating = true;
   captr->InUseCount++;

   captr->ValidatedCount++;

   OdsFileAcpInfo (&captr->FileOds, &CacheAcpInfoAst, rqptr); 
}

/*****************************************************************************/
/*
Called either explicitly or as an AST from an ACP QIO in CacheBegin(). Check if
modified, if not then just reply with a 302 header.  If contents should be
transfered to the client then begin. If there is a problem at all then declare
the AST to continue processing the request and let it worry about it, otherwise
begin providing the file data from the cache entry.
*/ 
 
CacheAcpInfoAst (REQUEST_STRUCT *rqptr)

{
   int  status;
   unsigned long  EndOfFileVbn;
   char  *cptr, *sptr, *zptr,
         *ContentTypePtr,
         *KeepAlivePtr;
   REQUEST_AST AstFunction;
   FILE_CENTRY  *captr, *tcaptr;
   FILE_CONTENT  *fcptr;
   LIST_ENTRY  *leptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheAcpInfoAst() !&F !&S !%D !%D", &CacheAcpInfoAst,
                 rqptr->rqCache.EntryPtr->FileOds.AcpInfo.AcpIOsb.Status,
                 &rqptr->rqCache.EntryPtr->RdtBinTime,
                 &rqptr->rqCache.EntryPtr->FileOds.AcpInfo.RdtBinTime);

   captr = rqptr->rqCache.EntryPtr;

   /* status of non-zero used to determine whether the ACP QIO was used */
   if (captr->FileOds.AcpInfo.AcpIOsb.Status)
   {
      /* finished getting the revalidate data */
      captr->CacheRevalidating = false;
      if (captr->InUseCount) captr->InUseCount--;

      /* first deassign the channel allocated by OdsFileAcpInfo() */
      sys$dassgn (captr->FileOds.AcpInfo.Channel);

      if (VMSnok (captr->FileOds.AcpInfo.AcpIOsb.Status))
      {
         /***********************/
         /* file access problem */
         /***********************/

         /* no longer valid data */
         CacheEntryNotValid (captr);

         /* move it to the end of the cache list ready for reuse */
         ListMoveTail (&CacheList, captr);

         /* continue processing the request */
         rqptr->rqCache.EntryPtr = NULL;
         FileEnd (rqptr);
         return;
      }

      EndOfFileVbn = captr->FileOds.AcpInfo.RecAttr.EndOfFileVbnLo +
                     (captr->FileOds.AcpInfo.RecAttr.EndOfFileVbnHi << 16);

      if (captr->RdtBinTime[0] !=
          captr->FileOds.AcpInfo.RdtBinTime[0] ||
          captr->RdtBinTime[1] !=
          captr->FileOds.AcpInfo.RdtBinTime[1] ||
          captr->EndOfFileVbn != EndOfFileVbn ||
          captr->FirstFreeByte != captr->FileOds.AcpInfo.RecAttr.FirstFreeByte)
      {
         /************************/
         /* data no longer valid */
         /************************/

         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "CACHE stale !AZ !%D (!%D / !%D) (!UL,!UL / !UL,!UL)",
                       captr->FileOds.ExpFileName,
                       &captr->CdtBinTime,
                       &captr->RdtBinTime,
                       &captr->FileOds.AcpInfo.RdtBinTime,
                       captr->EndOfFileVbn,
                       captr->FirstFreeByte,
                       EndOfFileVbn,
                       captr->FileOds.AcpInfo.RecAttr.FirstFreeByte);

         /* no longer valid data */
         CacheEntryNotValid (captr);

         /* move it to the end of the cache list ready for reuse */
         ListMoveTail (&CacheList, captr);

         /* continue processing the request */
         rqptr->rqCache.EntryPtr = NULL;
         FileCacheStale (rqptr);
         return;
      }

      /********************/
      /* data still valid */ 
      /********************/

      /* note the time the cached data was last validated */
      memcpy (&captr->ValidateBinTime, &rqptr->rqTime.Vms64bit, 8);
      captr->ValidateTickSecond = HttpdTickSecond + CacheValidateSeconds;

      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE revalidate !AZ", captr->FileOds.ExpFileName);
   }

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "CACHE hit !&?permanent\rvolatile\r !AZ",
                 captr->CachePermanent, captr->FileOds.ExpFileName);

   /**************************************/
   /* request to be fulfilled from cache */
   /**************************************/

   /* cancel the no-such-file callback, we've got it! */
   rqptr->FileTaskPtr->NoSuchFileFunction = NULL;

   memcpy (&captr->HitBinTime, &rqptr->rqTime.Vms64bit, 8);
   captr->FrequentTickSecond = HttpdTickSecond + CacheFrequentSeconds;

   CacheHitCount++;
   if (!captr->HitCount) CacheHits0--;
   captr->HitCount++;
   if (captr->HitCount < 10)
      CacheHits10++;
   else
   if (captr->HitCount < 100)
      CacheHits100++;
   else
      CacheHits100plus++;

   rqptr->AccountingDone =
      InstanceGblSecIncrLong (&AccountingPtr->CacheHitCount);

   /* if this entry has more hits move it to the head of the cache list */
   leptr = CacheList.HeadPtr;
   tcaptr = (FILE_CENTRY*)leptr;
   if (captr->HitCount > tcaptr->HitCount)
      ListMoveHead (&CacheList, captr);
   else
   {
      /* if this entry has more hits than the one "above" it swap them */
      leptr = captr;
      if (leptr->PrevPtr)
      {
         leptr = leptr->PrevPtr;
         tcaptr = (FILE_CENTRY*)leptr;
         if (captr->HitCount > tcaptr->HitCount)
         {
            ListRemove (&CacheList, captr);
            ListAddBefore (&CacheList, leptr, captr);
         }
      }
   }

   if (captr->ContentHandlerFunction)
   {
      /***********************************/
      /* this file has a content handler */
      /***********************************/

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "!&X",
                    captr->ContentHandlerFunction);

      /* allow one extra character for a terminating null */
      rqptr->FileContentPtr = fcptr = (FILE_CONTENT*)
         VmGetHeap (rqptr, sizeof(FILE_CONTENT) + captr->ContentLength+1);
      fcptr->ContentSize = captr->ContentLength+1;

      /* buffer space immediately follows the structured storage */
      fcptr->ContentPtr = (char*)fcptr + sizeof(FILE_CONTENT);
      memcpy (fcptr->ContentPtr,
              captr->ContentPtr,
              fcptr->ContentLength = captr->ContentLength);
      /* we always terminate these buffers with a null - it's usually text! */
      fcptr->ContentPtr[fcptr->ContentLength] = '\0';

      /* populate the file contents structure with some file data */
      zptr = (sptr = fcptr->FileName) + sizeof(fcptr->FileName);
      for (cptr = captr->FileOds.ExpFileName;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         /* ok, so something went astray */
         rqptr->rqCache.EntryPtr = NULL;
         FileEnd (rqptr);
         return;
      }
      *sptr = '\0';
      fcptr->FileNameLength = sptr - fcptr->FileName;

      memcpy (&fcptr->CdtBinTime, &captr->CdtBinTime, 8);
      memcpy (&fcptr->RdtBinTime, &captr->RdtBinTime, 8);

      fcptr->UicGroup = captr->UicGroup;
      fcptr->UicMember = captr->UicMember;
      fcptr->Protection = captr->Protection;

      /* set the content structure handler so it gets control at FileEnd() */
      rqptr->FileContentPtr->ContentHandlerFunction =
         captr->ContentHandlerFunction;

      /* this cache entry is no longer associated with this request */
      rqptr->rqCache.EntryPtr = NULL;
      FileEnd (rqptr);
      return;
   }

   if (!rqptr->rqResponse.HeaderPtr)
   {
      /**************************/
      /* full response required */
      /**************************/

      if (rqptr->rqPathSet.CharsetPtr)
      {
         cptr = captr->FileOds.ResFileName;
         if (!*cptr) cptr = captr->FileOds.ExpFileName;
         rqptr->rqPathSet.CharsetPtr = FileSetCharset (rqptr, cptr);
      }

      status = FileResponseHeader (rqptr, captr->ContentTypePtr,
                                   captr->ContentLength, &captr->RdtBinTime);
      if (VMSnok (status))
      {
         if (status == LIB$_NEGTIM)
         {
            InstanceGblSecIncrLong (&AccountingPtr->CacheHitNotModifiedCount);
            captr->HitNotModifiedCount++;
         }
         /* this cache entry is no longer associated with this request */
         rqptr->rqCache.EntryPtr = NULL;
         FileEnd (rqptr);
         return;
      }

      /* quit here if the HTTP method was HEAD */
      if (rqptr->rqHeader.Method == HTTP_METHOD_HEAD)
      {
         /* this cache entry is no longer associated with this request */
         rqptr->rqCache.EntryPtr = NULL;
         FileEnd (rqptr);
         return;
      }
   }

   /* 'CacheInUse' keeps track of whether the entry is in use or not */
   captr->InUseCount++;

   rqptr->rqCache.ContentPtr = captr->ContentPtr;
   rqptr->rqCache.ContentLength = captr->ContentLength;

   /* network writes are checked for success, fudge the first one! */
   rqptr->rqNet.WriteIOsb.Status = SS$_NORMAL;

   /* begin the transfer */
   if (rqptr->rqOutput.BufferCount)
   {
      /* after ensuring the current contents are output */
      NetWriteFullFlush (rqptr, &CacheNext);
   }
   else
      CacheNext (rqptr);
}

/*****************************************************************************/
/*
Write the next (or first) block of data from the cache buffer to the client.
*/ 

CacheNext (REQUEST_STRUCT *rqptr)

{
   int  status,
        DataLength;
   unsigned char  *DataPtr;
   REQUEST_AST AstFunction;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "CacheNext() !&F !&S",
                 &CacheNext, rqptr->rqNet.WriteIOsb.Status);

   if (VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      /* network write has failed (as AST), bail out now */
      CacheEnd (rqptr);
      return;
   }

   if (rqptr->rqCache.ContentLength > OutputBufferSize)
   {
      DataPtr = rqptr->rqCache.ContentPtr;
      DataLength = OutputBufferSize;
      rqptr->rqCache.ContentPtr += DataLength;
      rqptr->rqCache.ContentLength -= DataLength;

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      {
         WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "CACHE !UL/!UL",
                    DataLength, rqptr->rqCache.ContentLength);
         WatchDataDump (DataPtr, DataLength);
      }

      NetWrite (rqptr, &CacheNext, DataPtr, DataLength);
      return;
   }
   else
   {
      DataPtr = rqptr->rqCache.ContentPtr;
      DataLength = rqptr->rqCache.ContentLength;
      rqptr->rqCache.ContentLength = 0;

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      {
         WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "CACHE !UL/!UL",
                    DataLength, rqptr->rqCache.ContentLength);
         WatchDataDump (DataPtr, DataLength);
      }

      NetWrite (rqptr, &CacheEnd, DataPtr, DataLength);
      return;
   }
}

/*****************************************************************************/
/*
End of transfer to client using cached contents.
*/

CacheEnd (REQUEST_STRUCT *rqptr)

{
   FILE_CENTRY  *captr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "CacheEnd() !&F", &CacheEnd);

   captr = rqptr->rqCache.EntryPtr;

   /* this cache entry is no longer in use (if now zero) */
   if (captr->InUseCount) captr->InUseCount--;

   if (captr->Purge && !captr->InUseCount)
      CachePurgeEntry (captr, captr->PurgeCompletely);

   /* this cache entry is no longer associated with this request */
   rqptr->rqCache.EntryPtr = NULL;
   FileEnd (rqptr);
} 

/*****************************************************************************/
/*
Find/create a cache structure ready to load a cache data buffer from a file. It
is entirely possible, given that the maximum number of cache entries has been
reached and all are currently in use (either for data transfer or being
loaded), that there will be no cache entry available for this request to use.
In this case the 'rqptr->rqCache.EntryPtr' will remain NULL indicating that.

NOTE: as with CacheSearch(), the path passed to this function MUST be the
mapped path, not the request path. With conditional mapping identical request
paths may be mapped to completely different virtual paths.

'SizeInBytes' is the APPROXIMATE and MAXIMUM size required for data storage.
The size of a VARIABLE record file can be used because the actual data content
is smaller (when the 16 bits for the record length is turned into 8 bits for a
new-line character!). Hence, with a VARIABLE record file the content length is
invariably a little less that the reported file size. The actual size of the
data in the cache entry is represented by 'ContentLength'. 
*/ 

CacheLoadBegin (REQUEST_STRUCT *rqptr)

{
   int  status,
        FileKBytesMax,
        HashValue,
        PathLength,
        SizeInBytes;
   char  *cptr, *sptr;
   FILE_CENTRY  *captr, *lcaptr;
   FILE_TASK  *ftkptr;
   LIST_ENTRY  *leptr;
   MD5_HASH  *md5ptr;

   /*********/
   /* begin */
   /*********/

   ftkptr = rqptr->FileTaskPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheLoadBegin() !16&H !UL !AZ !UL",
                 &ftkptr->Md5Hash, ftkptr->Md5Hash.HashLong[0] & 0xfff,
                 ftkptr->FileName, ftkptr->SizeInBytes);

   /* better check, caching could be enabled between ASTs in search and load */
   if (!CacheHashTableInitialised) CacheInit (false);

   /* if non-NULL 'rqptr->rqCache.EntryPtr' is used for buffer space */
   rqptr->rqCache.EntryPtr = NULL;

   if (rqptr->rqPathSet.CacheMaxKBytes)
      FileKBytesMax = rqptr->rqPathSet.CacheMaxKBytes;
   else
      FileKBytesMax = CacheFileKBytesMax;

   /* if it's larger than the maximum allowed then just forget it */
   if (((SizeInBytes = ftkptr->SizeInBytes) >> 10) > FileKBytesMax)
   {
      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE load !AZ fail, file too big (!UL>!UL)",
                    ftkptr->FileOds.ExpFileName,
                    SizeInBytes >> 10, FileKBytesMax);
      return;
   }

   /********************************/
   /* check it still doesn't exist */
   /********************************/

   /* check no-one else has already created the entry during AST processing */
   md5ptr = &rqptr->FileTaskPtr->Md5Hash;
   /* 12 bit hash value, 0..4095 from a fixed 3 bytes of the MD5 hash */
   HashValue = ftkptr->Md5Hash.HashLong[0] & 0xfff;
   for (captr = CacheHashTable[HashValue];
        captr;
        captr = captr->HashCollisionNextPtr)
   {
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
"valid:!&B loading:!&B revalidating:!&B !AZ !&?no-match\rmatch\r",
                    captr->CacheValid, captr->CacheLoading,
                    captr->CacheRevalidating, captr->FileOds.ExpFileName,
                    captr->Md5Hash.HashLong[0] != md5ptr->HashLong[0] ||
                    captr->Md5Hash.HashLong[1] != md5ptr->HashLong[1] ||
                    captr->Md5Hash.HashLong[2] != md5ptr->HashLong[2] ||
                    captr->Md5Hash.HashLong[3] != md5ptr->HashLong[3]);

      /* match each of the 4 sets of 4 bytes (longwords) in the MD5 hash */
      if (captr->Md5Hash.HashLong[0] != md5ptr->HashLong[0] ||
          captr->Md5Hash.HashLong[1] != md5ptr->HashLong[1] ||
          captr->Md5Hash.HashLong[2] != md5ptr->HashLong[2] ||
          captr->Md5Hash.HashLong[3] != md5ptr->HashLong[3])
         continue; 

      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE load !AZ fail, now exists!!",
                    ftkptr->FileOds.ExpFileName);
      return;
   }

   /************************/
   /* get some data buffer */
   /************************/

   /* calculate exactly how much data buffer we're going to use */
   if (SizeInBytes <= 512)
      SizeInBytes = 512;
   else
   {
      /* divide by whatever to get cache chunk */
      SizeInBytes = ((SizeInBytes / CacheChunkInBytes) + 1) * CacheChunkInBytes;
   }

   /*
      If we've run out of cache entries, or
      if we've run out of cache memory, or
      if there is an invalidated entry on the end of the list
         that is not otherwise unavailable.
   */
   if (CacheEntryCount >= CacheEntriesMax ||
       (CacheMemoryInUse + SizeInBytes) >> 10 > CacheTotalKBytesMax ||
       (CacheList.TailPtr &&
        !((FILE_CENTRY*)CacheList.TailPtr)->CacheValid &&
        !((FILE_CENTRY*)CacheList.TailPtr)->InUseCount))
   {
      /***********************/
      /* reuse a cache entry */
      /***********************/

      /* process the cache entry list from least to most recent */
      for (leptr = CacheList.TailPtr; leptr; leptr = leptr->PrevPtr)
      {
         captr = (FILE_CENTRY*)leptr;

         if (WATCHING(rqptr) &&
             WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
            WatchThis (NULL, FI_LI, WATCH_MOD_CACHE,
                       "!&Z !UL !UL !&B !&B !UL",
                       captr->FileOds.ExpFileName,
                       captr->CacheSize, captr->ContentLength,
                       captr->CacheValid, captr->CacheRevalidating,
                       captr->InUseCount);

         /* if it's permanent or in use in some way then just continue */
         if (captr->CachePermanent || captr->InUseCount) continue;

         /* if it can be considered frequently hit */
         if (captr->CacheValid &&
             CacheFrequentHits &&
             captr->HitCount > CacheFrequentHits &&
             captr->FrequentTickSecond > HttpdTickSecond) continue;

         /* note the number of entries loaded but never subsequently hit */
         if (captr->CacheValid && !captr->HitCount) CacheNoHitsCount++;

         /* no longer valid data */
         CacheEntryNotValid (captr);

         /* move it to the end of the cache list ready for reuse */
         ListMoveTail (&CacheList, captr);

         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "CACHE load !AZ reuse entry",
                       ftkptr->FileOds.ExpFileName);
         break;
      }

      /* if we got to the end of the list then we couldn't have found one! */
      if (!leptr) captr = NULL;
   }

   if (!captr)
   {
      if (CacheEntryCount < CacheEntriesMax)
      {
         /***************************************/
         /* allocate memory for new cache entry */
         /***************************************/

         captr = VmGetCache (sizeof(FILE_CENTRY));
         CacheMemoryInUse += sizeof(FILE_CENTRY);
         CacheEntryCount++;

         /* add it to the list */
         ListAddTail (&CacheList, captr);

         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "CACHE load new !&?permanent\rvolatile\r entry !AZ",
                       rqptr->rqPathSet.CachePermanent,
                       ftkptr->FileOds.ExpFileName);
      }
      else
      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE load !AZ fail, insufficient entries",
                    ftkptr->FileOds.ExpFileName);
   }


   if (WATCHING(rqptr) &&
       WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "!UL !&X", CacheEntryCount, captr);

   /**************************/
   /* cache entry available? */
   /**************************/

   /* couldn't reuse an entry, couldn't create a new one, hence no caching */
   if (!captr) return;

   /************************/
   /* allocate data buffer */
   /************************/

   if (rqptr->rqPathSet.CachePermanent)
   {
      /* permanent cache entry */
      captr->CachePermanent = true;
      SizeInBytes = ftkptr->SizeInBytes;
      if (SizeInBytes <= 64)
         SizeInBytes = 64;
      else
      {
         /* permanent cache chunk size is sixty-four bytes */
         SizeInBytes = ((SizeInBytes / 64) + 1) * 64;
      }
      if (captr->ContentPtr)
      {
         /* free any previously allocated volatile data memory */
         VmFreeCache (captr->ContentPtr, FI_LI);
         CacheMemoryInUse -= captr->CacheSize;
      }
      captr->ContentPtr = VmGetPermCache (captr->CacheSize = SizeInBytes);
      CachePermMemoryInUse += captr->CacheSize;
      CachePermEntryCount++;
      CacheEntryCount--;
   }
   else
   if (captr->ContentPtr &&
       captr->CacheSize != SizeInBytes)
   {
      VmFreeCache (captr->ContentPtr, FI_LI);
      CacheMemoryInUse -= captr->CacheSize;
      captr->ContentPtr = NULL;
      captr->CacheSize = captr->ContentBufferUsedCount = 0;
      captr->CacheScavenged = captr->CacheValid = false;
   }

   if (!captr->ContentPtr)
   {
      /* if we haven't got sufficient cache memory then free some entries */
      if ((CacheMemoryInUse + SizeInBytes) >> 10 > CacheTotalKBytesMax)
      {
         /*******************/
         /* scavenge memory */
         /*******************/

         CacheScavengeCount++;

         /* mark this one just so *it* won't be scavenged */
         captr->InUseCount++;

         /* process the cache entry list from least to most recent */
         for (leptr = CacheList.TailPtr; leptr; leptr = leptr->PrevPtr)
         {
            /* remember, use a separate pointer or we get very confused :^) */
            lcaptr = (FILE_CENTRY*)leptr;

            if (WATCHING(rqptr) &&
                WATCH_MODULE(WATCH_MOD_CACHE) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
               WatchThis (NULL, FI_LI, WATCH_MOD_CACHE,
                          "!&Z !UL !UL !&B !&B !UL",
                          captr->FileOds.ExpFileName,
                          captr->CacheSize, captr->ContentLength,
                          captr->CacheValid, captr->CacheRevalidating,
                          captr->InUseCount);

            /* if it's permanent or in use in some way then just continue */
            if (lcaptr->CachePermanent || lcaptr->InUseCount) continue;

            /* note the number of entries loaded but never subsequently hit */
            if (lcaptr->CacheValid && !lcaptr->HitCount) CacheNoHitsCount++;

            /* no longer valid data */
            CacheEntryNotValid (lcaptr);
            lcaptr->CacheScavenged = true;

            if (lcaptr->ContentPtr)
            {
               VmFreeCache (lcaptr->ContentPtr, FI_LI);
               CacheMemoryInUse -= lcaptr->CacheSize;
               lcaptr->ContentPtr = NULL;
               lcaptr->CacheSize = lcaptr->ContentBufferUsedCount = 0;
            }

            /* if we've scavenged enough then break */
            if ((CacheMemoryInUse + SizeInBytes) >> 10 < CacheTotalKBytesMax)
               break;
         }

         /* remove the scavenge prophylactic */
         captr->InUseCount--;

         /* if we still haven't got enough memory then forget it! */
         if ((CacheMemoryInUse + SizeInBytes) >> 10 > CacheTotalKBytesMax)
         {
            WriteFaoStdout (
"%!AZ-W-CACHE, !20%D, !UL scavenge failed, !UL (!UL) in use, !UL (!UL) max\n",
               Utility, 0, SizeInBytes, CacheMemoryInUse, CacheEntryCount,
               CacheTotalKBytesMax, CacheEntriesMax);

            CacheScavengeFailCount++;

            if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
               WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                          "CACHE load !AZ fail, memory scavenge",
                          ftkptr->FileOds.ExpFileName);

            /*************************/
            /* not caching this time */
            /*************************/

            return;
         }
      }

      /* allocate memory for the cache data */
      if (captr->CachePermanent)
      {
         captr->ContentPtr = VmGetPermCache (captr->CacheSize = SizeInBytes);
         CachePermMemoryInUse += captr->CacheSize;
      }
      else
      {
         captr->ContentPtr = VmGetCache (captr->CacheSize = SizeInBytes);
         CacheMemoryAllocateCount++;
         CacheMemoryInUse += captr->CacheSize;
      }
   }
   else
      CacheMemoryReuseCount++;

   captr->ContentBufferUsedCount++;

   if (WATCHING(rqptr) &&
       WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE, "!&X !UL",
                 captr->ContentPtr, SizeInBytes);

   /*********************/
   /* ready for loading */
   /*********************/

   /*
      By setting the path and hash table this entry can still be hit
      even though the data is still being loaded and is not yet valid.
      This will prevent the same path having concurrent, multiple loads.
   */

   memcpy (&captr->Md5Hash, &rqptr->FileTaskPtr->Md5Hash, sizeof(MD5_HASH));
   /* 12 bit hash value, 0..4095 from a fixed 3 bytes of the MD5 hash */
   HashValue = captr->Md5Hash.HashLong[0] & 0xfff;
   if (!CacheHashTable[HashValue])
   {
      /* set hash table index */
      CacheHashTable[HashValue] = captr;
      captr->HashCollisionPrevPtr = captr->HashCollisionNextPtr = NULL;
   }
   else
   {
      /* add to head of hash-collision list */
      lcaptr = CacheHashTable[HashValue];
      lcaptr->HashCollisionPrevPtr = captr;
      captr->HashCollisionPrevPtr = NULL;
      captr->HashCollisionNextPtr = lcaptr;
      CacheHashTable[HashValue] = captr;
   }

   /* copy the entire on-disk structure from file to cache */
   OdsCopyStructure (&captr->FileOds, &ftkptr->FileOds);

   /* the file size data */
   captr->EndOfFileVbn = ftkptr->EndOfFileVbn;
   captr->FirstFreeByte = ftkptr->FirstFreeByte;

   /* quadword time file was loaded/validated, created, last modified */
   memcpy (&captr->LoadBinTime, &rqptr->rqTime.Vms64bit, 8);
   memcpy (&captr->ValidateBinTime, &rqptr->rqTime.Vms64bit, 8);
   captr->ValidateTickSecond = HttpdTickSecond + CacheValidateSeconds;
   memcpy (&captr->CdtBinTime, &ftkptr->FileOds.XabDat.xab$q_cdt, 8);
   memcpy (&captr->RdtBinTime, &ftkptr->FileOds.XabDat.xab$q_rdt, 8);

   /* some file details */
   captr->UicGroup = ftkptr->FileOds.XabPro.xab$w_grp;
   captr->UicMember = ftkptr->FileOds.XabPro.xab$w_mbm;
   captr->Protection = ftkptr->FileOds.XabPro.xab$w_pro;

   captr->ContentTypePtr = ftkptr->ContentTypePtr;
   captr->ContentLength = captr->HitCount = 0;
   captr->InUseCount++;

   /* the cache entry will resuse the original content handler (if any) */
   captr->ContentHandlerFunction = ftkptr->ContentHandlerFunction;

   captr->CacheScavenged = captr->CacheValid = captr->CacheLoadValid = false;
   captr->CacheLoading = true;

   /* point the request at this cache entry */
   rqptr->rqCache.EntryPtr = captr;
}

/*****************************************************************************/
/*
End loading a cache data buffer from a file.
*/ 

CacheLoadEnd (REQUEST_STRUCT *rqptr)

{
   int  Length;
   FILE_CENTRY  *captr;

   /*********/
   /* begin */
   /*********/

   captr = rqptr->rqCache.EntryPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheLoadEnd() !UL", captr->CacheLoadValid);

   if (captr->CacheValid = captr->CacheLoadValid)
   {
      InstanceGblSecIncrLong (&AccountingPtr->CacheLoadCount);
      CacheLoadCount++;
      CacheHits0++;

      /* move entry to the head of the cache list */
      ListMoveHead (&CacheList, captr);

      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE load !AZ complete (!UL bytes)",
                     captr->FileOds.ExpFileName, captr->ContentLength);
   }
   else
   {
      /* remove any vestiges of the failed load */
      CacheEntryNotValid (captr);

      /* move it to the end of the cache list ready for reuse */
      ListMoveTail (&CacheList, captr);

      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "CACHE load !AZ failed, not valid",
                     captr->FileOds.ExpFileName);
   }

   /* obviously no longer loading! */
   captr->CacheLoading = false;

   /* this cache entry is no longer in use (if now zero) */
   if (captr->InUseCount) captr->InUseCount--;

   /* this cache entry is no longer associated this request */
   rqptr->rqCache.EntryPtr = NULL;
}

/*****************************************************************************/
/*
Scan through the cache list.  If a cache entry is currently not in use then
free the data memory associated with it.  If purge completely then also remove
the entry from the list and free it's memory.  If the entry is currently in use
then mark it for purge and if necessary for complete removal.
*/

CachePurge
(
BOOL Completely,
int *PurgeCountPtr,
int *MarkedForPurgeCountPtr
)
{
   int  PurgeCount,
        MarkedForPurgeCount;
   FILE_CENTRY  *captr;
   LIST_ENTRY  *leptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "CachePurge() !UL", Completely);

   CacheZeroCounters ();

   PurgeCount = MarkedForPurgeCount = 0;

   /* do it backwards! seeing they are pushed to the tail of the list */
   leptr = CacheList.TailPtr;
   while (leptr)
   {
      captr = (FILE_CENTRY*)leptr;

      /* now, before stuffing around with the entry get the next one */
      leptr = leptr->PrevPtr;

      if (WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_CACHE,
                    "!&Z !UL !UL !&B !&B !UL",
                    captr->FileOds.ExpFileName,
                    captr->CacheSize, captr->ContentLength,
                    captr->CacheValid, captr->CacheRevalidating,
                    captr->InUseCount);

      if (captr->InUseCount)
      {
         if (Completely)
            captr->Purge = captr->PurgeCompletely = true;
         else
            captr->Purge = true;
         MarkedForPurgeCount++;
      }
      else
      {
         if (captr->CacheSize)
         {
            CachePurgeEntry (captr, Completely);
            PurgeCount++;
         }
      }
   }

   if (PurgeCountPtr) *PurgeCountPtr = PurgeCount;
   if (MarkedForPurgeCountPtr)
      *MarkedForPurgeCountPtr = MarkedForPurgeCount;
} 

/*****************************************************************************/
/*
Purge a cache entry, either just the data, or completely from the cache list.
*/

CachePurgeEntry
(
FILE_CENTRY *captr,
BOOL Completely
)
{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (NULL, FI_LI, WATCH_MOD_CACHE,
                 "CachePurgeEntry() !UL", Completely);

   if (captr->InUseCount)
   {
      if (Completely)
         captr->Purge = captr->PurgeCompletely = true;
      else
         captr->Purge = true;
      return;
   }

   CacheEntryNotValid (captr);

   if (captr->ContentPtr)
   {
      /* CacheEntryNotValid() has already removed a permanent entry's data */ 
      VmFreeCache (captr->ContentPtr, FI_LI);
      CacheMemoryInUse -= captr->CacheSize;
      captr->ContentPtr = NULL;
   }

   if (Completely)
   {
      ListRemove (&CacheList, captr);
      VmFreeCache (captr, FI_LI);
      CacheMemoryInUse -= sizeof(FILE_CENTRY);
   }
   else
   {
      /* move it to the end of the cache list ready for reuse */
      ListMoveTail (&CacheList, captr);
      captr->ContentLength = captr->CacheSize = 0;
      captr->CacheValid = captr->Purge = false;
   }
} 

/*****************************************************************************/
/*
Remove from hash table and any associated collision list.
*/ 

#ifdef __DECC
#pragma inline(CacheEntryNotValid)
#endif

CacheEntryNotValid (FILE_CENTRY *captr)

{
   int  HashValue;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "CacheEntryNotValid()");

   /* 12 bit hash value, 0..4095 from a fixed 3 bytes of the MD5 hash */
   HashValue = captr->Md5Hash.HashLong[0] & 0xfff;
   if ((FILE_CENTRY*)(CacheHashTable[HashValue]) == captr)
   {
      /* must be at the head of any collision list */
      CacheHashTable[HashValue] = captr->HashCollisionNextPtr;
      if (captr->HashCollisionNextPtr)
         captr->HashCollisionNextPtr->HashCollisionPrevPtr =
            captr->HashCollisionPrevPtr;
   }
   else
   {
      /* if somewhere along the collision list */
      if (captr->HashCollisionPrevPtr)
         captr->HashCollisionPrevPtr->HashCollisionNextPtr =
            captr->HashCollisionNextPtr;
      /*
         *** SITE OF ONE OF MY MOST STUPID AND COSTLY PROGRAMMING ERRORS ***
         If this isn't an argument against coding for speed and efficiency
         instead of with tested and common code routines I don't know what is!
      */
      if (captr->HashCollisionNextPtr)
         captr->HashCollisionNextPtr->HashCollisionPrevPtr =
            captr->HashCollisionPrevPtr;
   }

   captr->CacheValid = false;
   captr->HashCollisionPrevPtr = captr->HashCollisionNextPtr = NULL;

   if (captr->CachePermanent)
   {
      /* during cache purge or when a request is aborted during load */
      captr->CachePermanent = false;
      if (captr->ContentPtr)
      {
         VmFreePermCache (captr->ContentPtr, FI_LI);
         CachePermMemoryInUse -= captr->CacheSize;
         captr->ContentPtr = NULL;
         captr->CacheSize = 0;
      }
      CacheEntryCount++;
      CachePermEntryCount--;
   }
}

/*****************************************************************************/
/*
Return a report on cache usage.  This function blocks while executing.
*/ 

CacheReport
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
BOOL IncludeEntries
)
{
   static char BeginPageFao [] =
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Configuration</TH><TH>Activity</TH><TH>Files</TH></TR>\n\
\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Caching:</TH>\
<TD ALIGN=left COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Memory&nbsp;&nbsp;/Permanent:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>kB</FONT></TD></TR>\n\
<TR><TH ALIGN=right>/Volatile:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>kB</FONT></TD></TR>\n\
<TR><TH ALIGN=right>/Max:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>kB</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Entries&nbsp;&nbsp;/Permanent:</TH>\
<TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>/Volatile:</TH>\
<TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>/Max:</TH>\
<TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Hash Table:</TH>\
<TD ALIGN=right>!UL</TD>\
<TD><FONT SIZE=-1>entries</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max File Size:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>kB</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Chunk Size:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>kB</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Validate:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>seconds</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Frequent&nbsp;&nbsp;/Hits:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>hits</FONT></TD></TR>\n\
<TR><TH ALIGN=right>/Within:</TH>\
<TD ALIGN=right>!UL</TD><TD><FONT SIZE=-1>seconds</FONT></TD></TR>\n\
</TABLE>\n\
\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Search:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right><U>Hash Table</U></TH></TR>\n\
<TR><TH ALIGN=right>Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Miss:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right><U>Collision</U></TH></TR>\n\
<TR><TH ALIGN=right>Total:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Miss:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right><U>Load Memory</U></TH></TR>\n\
<TR><TH ALIGN=right>Allocate:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Reuse:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Scavenge:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Scavenge Fail:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
</TABLE>\n\
\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Load:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Not Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Total Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>1-9:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>10-99:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>>100:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
</TABLE>\n\
\
</TD></TR>\n\
</TABLE>\n\
!&@";

   static char  EntriesFao [] =
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=12>Entries</TH></TR>\
<TR><TH ROWSPAN=3></TH><TH COLSPAN=11>File Name</TH></TR>\
<TR>\
<TH>Type</TH><TH>Size</TH><TH>Length</TH><TH>In-Use</TH><TH>Hash</TH>\
<TH>Revised</TH><TH COLSPAN=2>Validated</TH>\
<TH>Loaded</TH><TH COLSPAN=2>Hit&nbsp;/&nbsp;304</TH>\
</TR>\n\
<TR></TR>\n";

   static char  CacheFao [] =
"<TR><TH ROWSPAN=2 ALIGN=right>!UL&nbsp;</TH>\
<TD COLSPAN=11><TT>!AZ</TT></TR>\n\
<TR>\
<TD ALIGN=center>!&?P\rV\r</TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL&nbsp;/&nbsp;!UL</TD>\
<TD><FONT SIZE=1>!20&W</FONT></TD>\
<TD>!&@</TD><TD ALIGN=right>!UL</TD>\
<TD><FONT SIZE=1>!20&W</FONT></TD>\
<TD><FONT SIZE=1>!20&W</FONT></TD>\
<TD ALIGN=right>!UL&nbsp;/&nbsp;!UL</TD>\
</TR>\n";

   static char  EntriesButtonFao [] =
"</TABLE>\n\
<P><FONT SIZE=+1>[<A HREF=\"!AZ\">Entries</A>]</FONT>\n\
</BODY>\n\
</HTML>\n";

   static char  EndPageFao [] =
"</TABLE>\n\
</BODY>\n\
</HTML>\n";

   int  cnt,
        status,
        Count,
        HashCollisionListLength;
   unsigned long  FaoVector [64];
   unsigned long  *vecptr;
   FILE_CENTRY  *captr;
   LIST_ENTRY  *leptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_CACHE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_CACHE,
                 "CacheReport() !&A", NextTaskFunction);

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   RESPONSE_HEADER_200_HTML (rqptr);
   AdminPageTitle (rqptr, "Cache Report");

   vecptr = FaoVector;

   if (CacheEnabled)
      *vecptr++ = "[enabled]";
   else
      *vecptr++ = "<FONT COLOR=\"#ff000\">[disabled]</FONT>";

   *vecptr++ = CachePermMemoryInUse >> 10;
   *vecptr++ = CacheMemoryInUse >> 10;
   *vecptr++ = CacheTotalKBytesMax;
   *vecptr++ = CachePermEntryCount;
   *vecptr++ = CacheEntryCount;
   *vecptr++ = CacheEntriesMax;
   *vecptr++ = CACHE_HASH_TABLE_ENTRIES;
   *vecptr++ = CacheFileKBytesMax;
   *vecptr++ = CacheChunkInBytes >> 10;
   *vecptr++ = CacheValidateSeconds;
   *vecptr++ = CacheFrequentHits;
   *vecptr++ = CacheFrequentSeconds;

   *vecptr++ = CacheHashTableCount;
   *vecptr++ = CacheHashTableHitCount;
   if (CacheHashTableCount)
      *vecptr++ = CacheHashTableHitCount * 100 / CacheHashTableCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHashTableMissCount;
   if (CacheHashTableCount)
      *vecptr++ = CacheHashTableMissCount * 100 / CacheHashTableCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHashTableCollsnCount;
   if (CacheHashTableCount)
      *vecptr++ = CacheHashTableCollsnCount * 100 / CacheHashTableCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHashTableCollsnHitCount;
   if (CacheHashTableCount)
      *vecptr++ = CacheHashTableCollsnHitCount * 100 / CacheHashTableCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHashTableCollsnMissCount;
   if (CacheHashTableCount)
      *vecptr++ = CacheHashTableCollsnMissCount * 100 / CacheHashTableCount;
   else
      *vecptr++ = 0;

   *vecptr++ = CacheMemoryAllocateCount;
   if (CacheLoadCount)
      *vecptr++ = CacheMemoryAllocateCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheMemoryReuseCount;
   if (CacheLoadCount)
      *vecptr++ = CacheMemoryReuseCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheScavengeCount;
   if (CacheLoadCount)
      *vecptr++ = CacheScavengeCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheScavengeFailCount;
   if (CacheLoadCount)
      *vecptr++ = CacheScavengeFailCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;

   *vecptr++ = CacheLoadCount;
   *vecptr++ = CacheHits0;
   if (CacheLoadCount)
      *vecptr++ = CacheHits0 * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHitCount;
   if (CacheLoadCount)
      *vecptr++ = CacheHitCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits10;
   if (CacheHitCount)
      *vecptr++ = CacheHits10 * 100 / CacheHitCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits100;
   if (CacheHitCount)
      *vecptr++ = CacheHits100 * 100 / CacheHitCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits100plus;
   if (CacheHitCount)
      *vecptr++ = CacheHits100plus * 100 / CacheHitCount;
   else
      *vecptr++ = 0;

   if (!IncludeEntries)
   {
      *vecptr++ = EntriesButtonFao;
      *vecptr++ = ADMIN_REPORT_CACHE_ENTRIES;
   }
   else
      *vecptr++ = "";

   status = NetWriteFaol (rqptr, BeginPageFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   if (!IncludeEntries)
   {
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /*****************/
   /* cache entries */
   /*****************/

   status = NetWriteFaol (rqptr, EntriesFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   Count = 0;

   if (WATCHING(rqptr) &&
       WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
       WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "!&X !&X",
                  CacheList.HeadPtr, CacheList.TailPtr);

   /* process the cache entry list from most to least recently hit */
   for (leptr = CacheList.HeadPtr; leptr; leptr = leptr->NextPtr)
   {
      captr = (FILE_CENTRY*)leptr;

      if (WATCHING(rqptr) &&
          WATCH_MODULE(WATCH_MOD_CACHE) && WATCH_MODULE(WATCH_MOD__DETAIL))
          WatchThis (NULL, FI_LI, WATCH_MOD_CACHE, "!&X<-!&X->!&X !&X",
                     leptr->PrevPtr, leptr, leptr->NextPtr, captr);

      HashCollisionListLength = 0;
      /* count further down the list */
      while (captr->HashCollisionNextPtr)
      {
         captr = captr->HashCollisionNextPtr;
         HashCollisionListLength++;
      }
      captr = (FILE_CENTRY*)leptr;
      /* count further up the list */
      while (captr->HashCollisionPrevPtr)
      {
         captr = captr->HashCollisionPrevPtr;
         HashCollisionListLength++;
      }
      captr = (FILE_CENTRY*)leptr;

      vecptr = FaoVector;
      *vecptr++ = ++Count;
      *vecptr++ = captr->FileOds.ExpFileName;
      *vecptr++ = captr->CachePermanent;
      *vecptr++ = captr->CacheSize;
      *vecptr++ = captr->ContentLength;
      *vecptr++ = captr->InUseCount;
      /* 12 bit hash value, 0..4095 from a fixed 3 bytes of the MD5 hash */
      *vecptr++ = captr->Md5Hash.HashLong[0] & 0xfff;
      *vecptr++ = HashCollisionListLength;
      *vecptr++ = &captr->RdtBinTime;

      if (captr->CacheValid)
      {
         *vecptr++ = "<FONT SIZE=1>!20&W</FONT>";
         *vecptr++ = &captr->ValidateBinTime;
      }
      else
      if (captr->CacheScavenged)
         *vecptr++ = "<B>SCAVENGED</B>";
      else
      if (captr->CacheLoading)
         *vecptr++ = "<B>LOADING</B>";
      else
         *vecptr++ = "<B>INVALID</B>";

      *vecptr++ = captr->ValidatedCount;
      *vecptr++ = &captr->LoadBinTime;
      *vecptr++ = &captr->HitBinTime;
      *vecptr++ = captr->HitCount;
      *vecptr++ = captr->HitNotModifiedCount;

      status = NetWriteFaol (rqptr, CacheFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   status = NetWriteFaol (rqptr, EndPageFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/

