/*****************************************************************************/
/*
                                 File.c

This module implements a full multi-threaded, AST-driven, asynchronous file 
send.  The AST-driven nature makes the code a little more difficult to follow, 
but creates a powerful, event-driven, multi-threaded server.  All of the 
necessary functions implementing this module are designed to be non-blocking. 

It can operate in one of four modes.

1) File direct to network.
   ~~~~~~~~~~~~~~~~~~~~~~
A smallish, standard output buffer is allocated if variable length record file,
successive records are read into the buffer until it fills, then written to the
network as a single block.  Records always have a newline character added to
each record (variable length record files are invariably text).  If a stream,
fixed or undefined record format the buffer is filled in a single virtual block
read and then immediately written to the network.

It uses the same buffer space as, and interworks with, NetWriteBuffered().  If 
there is already data (text) buffered in this area the file module will, for 
record-oriented, non-HTML-escaped transfers, continue to fill the area (using 
its own buffering function), flushing when and if necessary.  At end-of-file 
explicitly flush the buffer only if escaping HTML-forbidden characters, 
otherwise allow subsequent processing to do it as necessary.  For block-mode 
files the buffer is explicitly flushed before commencing the file transfer. 

2) File into file cache buffer, simultaneously to network.
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A file cache buffer of sufficient size to contain the whole file is allocated. 
If reading in record mode this is filled with successive reads (as per above)
until a chunk the size of a standard output buffer is filled, at which time it
is written to the network as a block.  In block mode each group of blocks read
is output to the network.

3) File into contents buffer.
   ~~~~~~~~~~~~~~~~~~~~~~~~~
A file contents buffer of sufficient size to contain the whole file is
allocated.  When reading in either record or block mode the buffer is just
filled with the file reads - NO NETWORK OUTPUT.  When complete the request gets
control of the contents buffer for subsequent processing.

4) File into file cache as well as into contents buffer.
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Both file cache and contents buffers of sufficient size are allocated.  Records
and blocks fill both buffers during the reads - NO NETWORK OUTPUT.  When
complete the file is cached and the contents buffer filled.  When complete the
request gets control of the contents buffer for subsequent processing. 
(Subsequent access to the file can then be serviced from the file cache.)


MD5 DIGEST
----------
An MD5 digest (16 byte hash) is used to uniquely identify cached files.  The
hash is generated either from a mapped path or from the file name.  See the
CACHE.C module for further detail.


FILE LANGUAGE VARIANTS
----------------------
If the path is SET to ACCEPT=LANG then this module attempts to find
language-specific variants of the file.  The format of the file names for these
variants is <file-name>.<file-type>_<language> where 'language' is one of the
ISO language abbreviations, e.g. "en" for English, "fr" for French, "de" for
German, "ru" for Russian, etc.  Hence if the basic file name is EXAMPLE.HTML
then a specifically English version would be EXAMPLE.HTML_EN, a French version
EXAMPLE.HTML_FR, etc.  Language variants may be provided for any file type the
HTTPD$CONFIG directive [AddType] specifies as content-type "text/..".

The language variant code behaves as follows.  When FileBegin() is called with
a path SET to ACCEPT=LANG and a default language is specified (for those files
without the language variant abbreviation) this is checked against the
request's accepted languages to see if the default would be the request's first
choice.  If so then there is no need for further accept language processing. 
If not then a series of functions search using the basic specification for
files matching "EXAMPLE.HTML_*".  All files matching this wildcard have the '*'
portion (e.g. "EN", "FR", "DE", "RU") added to a list of matching variants. 
When the search is complete a final function compares each of the request's
"Accept-Language:" list to each of the searched-for variants.  If one matches
the contents of that file are returned.  If none are matched the original
EXAMPLE.HTML would be returned.

Example behaviour.  A directory contains

  EXAMPLE.HTML
  EXAMPLE.HTML_FR
  EXAMPLE.HTML_DE
  EXAMPLE.HTML_RU

and a request specifies

  Accept-Language: en,fr,de

then the EXAMPLE.HTML_FR file will be returned.  For a request specifying

  Accept-Language: ru

then the EXAMPLE.HTML_RU file is returned, and if

  Accept-Language: en

EXAMPLE.HTML returned (without having searched for any other variants).

One or other file is always returned, with the default, non-language file
always the fallback source of data.  If it does not exist and no other
language variant is selected the request returns a 404 file-not-found error.


GENERAL COMMENTS
----------------
Module handles both VARIABLE file (record-by-record) and STREAM, FIXED and
UNDEFINED file (virtual block- oriented) transfers. Even with record mode
efficiency is increased by building up a series of file records in available
buffer space before transfering them as one I/O to the network. The binary mode
reads as many virtual blocks as will fit into the available buffer, then
transfers these as one I/O to the network. NON-VARIABLE FORMAT FILES ARE BY FAR
THE MOST EFFICIENT TO HANDLE.

For text mode transfers (record-oriented) the completion of each read of the 
next record using FileNextRecord() calls FileNextRecordAst().  This function 
checks if the record buffer is full (by detecting if the last read failed with 
a "record too big for user's buffer" error).  If full it writes the buffer to 
the network using NetWrite() or other buffering action (for cache and
contents), which when complete calls FileNextRecord() as an AST routine.  This
begins by getting the next record again (the one that failed).  If the buffer
is not full FileNextRecordAst() just calls  FileNextRecord() to get the next
record, until the buffer fills. 

For binary mode transfers (non-record-oriented) the completion of each read of 
the next series of blocks using FileNextBlocks() calls FileNextBlocksAst().  
This function writes the blocks (binary) using NetWrite() or other buffering
action, which when complete calls FileNextBlocks() as an AST.  This in turn
calls  FileNextBlocksAst(), etc., which drives the transfer until EOF. 

When not buffering to cache or contents the module can encapsulate plain-text
and escape HTML-forbidden characters.

Works in conjunction with the CACHE module.  A file read can be simultaneously
used to send the data to the client and load a cache buffer. The request
structure fields 'CacheContentPtr' and 'CacheContentCount' being used in both
block I/O and record mode access to track through the available buffer space,
for this located in the cache structure. For non-cache-load reads uses the
standard buffer space pointed to by 'OutputBufferPtr' and for record mode
access tracked using 'OutputBufferCurrentPtr' and 'OutputBufferRemaining'
fields.

Implements defacto HTTP/1.0 persistent connections.  Only provide "keep-alive"
response if we're sending a file in binary mode and know its precise length.
An accurate "Content-Length:" field is vital for the client's correct reception
of the transmitted data.  Currently the only other time a "keep-alive" response
can be provided is when a "304 not-modified" header is returned (it has a
content length of precisely zero!).  October 1997: noted that Netscape
Navigator 3.n seems to pay no attention to a "Content-Length: 0" with a
"Keep-Alive:" connection, it just sits there waiting until the keep-alive time
closes the connection, after which it reports "document contains no data".
(IE 3.02 seems to behave correctly ;^)  I have therefore disabled persistent
connections for zero-length files.

Uses the ACP-QIO interface detailed in the "OpenVMS I/O User's Reference
Manual" to retrieve file record attributes, revision date/time and size.


VERSION HISTORY
---------------
12-AUG-2003  MGD  access to HTA or HTL file now reports not found
16-JUN-2003  MGD  bugfix; FileSetCharset() following initial CacheSearch()
                  moved to CACHE.C module (ACCVIO if entry NULLed)
05-OCT-2002  MGD  no sneaky getting directory contents by downloading files!
                  refine VMS security profile usage
03-JUN-2002  MGD  bugfix; (well sort of) it would appear that after NO_CONCEAL
                  searching and a sys$open() you must sys$close() *before* the
                  SYNCHCK sys$parse() release of resources otherwise a channel
                  to the device is left assigned!!
27-APR-2002  MGD  make SYSPRV enabled ASTs asynchronous
19-MAR-2002  MGD  bugfix; OdsParse() for VMS authenticated request
15-MAR-2002  MGD  bugfix; FileNextRecordAst() VAR file into contents buffer
18-NOV-2001  MGD  FileAcceptLang..()
23-OCT-2001  MGD  bugfix; FileNextBlocksAst() 'ContentRemaining'
04-AUG-2001  MGD  modifications in line with changes in the handling
                  of file and cache (now MD5 hash based) processing,
                  block I/O complete if _rsz is less than _usz
                  support module WATCHing
10-MAY-2001  MGD  bugfix; FileNextRecordAst() buffer flush
05-MAR-2001  MGD  bugfix; FileNextBlocks() 32767 to 0xfe00 (65024)
29-DEC-2000  MGD  allow a file's contents to be read into a buffer
23-DEC-2000  MGD  allow access to an HTL if it is authorized
06-DEC-2000  MGD  make a search list DNF appear as a FNF
01-SEP-2000  MGD  add optional, local path authorization
                  (for calls from the likes of SSI.C)
09-JUN-2000  MGD  search-list processing refined
04-MAR-2000  MGD  use NetWriteInit(), et.al.
27-DEC-1999  MGD  support ODS-2 and ODS-5 using ODS module
10-OCT-1999  MGD  "scrunched" (in fact all) SSI files, prevent streamLFing
17-SEP-1999  MGD  bugfix; sys$parse() NAM$M_NOCONCEAL for search lists
23-MAY-1999  MGD  do not allow "?httpd=content" requests to be cached
05-FEB-1999  MGD  bugfix; FileNextRecord() zero '_usz'
07-NOV-1998  MGD  WATCH facility
19-SEP-1998  MGD  improve granularity of file open, connect, close, ACP
14-MAY-1998  MGD  request-specified content-type ("httpd=content&type=")
19-MAR-1998  MGD  buffer VBN and first free byte for use by cache module
19-JAN-1998  MGD  new NetWriteBuffered() and structures
07-JAN-1998  MGD  groan, bugfix; record processing for files > 4096 bytes
                  completely brain-dead ... sorry
22-NOV-1997  MGD  sigh, bugfix; heap corruption by file cache
05-OCT-1997  MGD  file cache,
                  keep-alive now not used if the content-length is zero
17-SEP-1997  MGD  if block-mode open is locked retry open in record-mode
17-AUG-1997  MGD  message database,
                  SYSUAF-authenticated users security-profile,
                  addressed potential problem with FIXed and odd-byte records
08-JUN-1997  MGD  if request "Pragma: no-cache" then always return 
27-FEB-1997  MGD  delete on close for "temporary" files
01-FEB-1997  MGD  HTTPd version 4
01-AUG-1996  MGD  Variable to stream-LF file conversion "on-the-fly"
12-APR-1996  MGD  determine read method (record or binary) from record format;
                  implemented persistent connections ("keep-alive")
01-DEC-1995  MGD  HTTPd version 3
27-SEP-1995  MGD  added If-Modified-Since: functionality;
                  changed carriage-control on records from <CR><LF> to single
                  <LF> ('\n' ... newline), to better comply with some browsers
                  (Netscape was spitting on X-bitmap files, for example!)
07-AUG-1995  MGD  ConfigcfReport.MetaInfoEnabled to allow physical file
                  specification included as commentary within an HTML file
13-JUL-1995  MGD  bugfix; occasionally a record was re-read after flushing
                  the records accumulated in the buffer NOT due to RMS$_RTB
20-DEC-1994  MGD  initial development for multi-threaded daemon
*/
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

/* standard C header files */
#include <ctype.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <atrdef.h>
#include <descrip.h>
#include <iodef.h>
#include <libdef.h>
#include <rms.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "ods.h"
#include "wasd.h"

#define WASD_MODULE "FILE"

/******************/
/* global storage */
/******************/

char ErrorBufferFileMaxBytes [] =
  "Exceeds reasonable limit on the size of this type of file.";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern BOOL Debug;
#else
#define Debug 0 
#endif

extern BOOL  CacheEnabled,
             OdsExtended;

extern int  OutputBufferSize;

extern unsigned long  SysPrvMask[];

extern char  ConfigContentTypeSsi[],
             ConfigContentTypeUnknown[],
             ConfigDefaultFileContentType[],
             ErrorSanityCheck[],
             HttpProtocol[],
             SoftwareID[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern MSG_STRUCT  Msgs;
extern WATCH_STRUCT  Watch;

/*****************************************************************************/
/*
Initalize the file task structure.
NOTE: FILE PATHS ARE ALWAYS AUTHORIZED UNLESS EXPLICITLY TURNED OFF.
*/ 
 
FILE_TASK* FileNewTask (REQUEST_STRUCT *rqptr)

{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileNewTask()");

   if ((tkptr = rqptr->FileTaskPtr) && tkptr->TaskInitialized)
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   /* set up the task structure (possibly multiple serially) */
   if (!(tkptr = rqptr->FileTaskPtr))
      rqptr->FileTaskPtr = tkptr =
         (FILE_TASK*)VmGetHeap (rqptr, sizeof(FILE_TASK));
   else
   if (!tkptr->TaskInitialized)
      memset (tkptr, 0, sizeof(FILE_TASK));

   tkptr->TaskInitialized = true;
   tkptr->AuthorizePath = true;

   return (tkptr);
}

/*****************************************************************************/
/*
Authorize the path to FileBegin() 'FileName' before accessing.
NOTE: FILE PATHS ARE ALWAYS AUTHORIZED UNLESS EXPLICITLY TURNED OFF.
*/ 
 
void FileSetAuthorizePath
(
REQUEST_STRUCT *rqptr,
BOOL YesNo
)
{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetAuthorizePath() !&B", YesNo);

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   tkptr->AuthorizePath = YesNo;
}

/*****************************************************************************/
/*
Read the file into memory pointed to by 'rqptr->FileContentPtr->ContentPtr'
and 'rqptr->FileContentPtr->ContentLength' in length.  File is null-terminated
in case it is text of some sort (this null is not counted in the length).
*/ 
 
void FileSetContentHandler
(
REQUEST_STRUCT *rqptr,
REQUEST_AST ContentHandlerFunction,
int SizeMax
)
{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetContentHandler() !&X !UL",
                 ContentHandlerFunction, SizeMax);

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   /* setting this non-NULL indicates the file contents should be buffered */
   tkptr->ContentHandlerFunction = ContentHandlerFunction;
   tkptr->FileContentsSizeMax = SizeMax;
}

/*****************************************************************************/
/*
File allowed to be cached?
*/ 
 
void FileSetCacheAllowed
(
REQUEST_STRUCT *rqptr,
BOOL YesNo
)
{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetCacheAllowed() !&B", YesNo);

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   tkptr->CacheAllowed = YesNo;
}

/*****************************************************************************/
/*
Escape HTML-forbidden characters (e.g. '<') during file output.
*/ 
 
void FileSetEscapeHtml
(
REQUEST_STRUCT *rqptr,
BOOL YesNo
)
{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetEscapeHtml() !&B", YesNo);

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   tkptr->EscapeHtml = YesNo;
   if (!rqptr->rqCache.DoNotCache) rqptr->rqCache.DoNotCache = YesNo;
}

/*****************************************************************************/
/*
Enclose the file output with <PRE>...</PRE> tags.
*/ 
 
void FileSetPreTag
(
REQUEST_STRUCT *rqptr,
BOOL YesNo
)
{
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetPreTag() !&B", YesNo);

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   tkptr->PreTagFileContents = YesNo;
   if (!rqptr->rqCache.DoNotCache) rqptr->rqCache.DoNotCache = YesNo;
}

/*****************************************************************************/
/*
Begin to transfer a file.  
*/ 
 
FileBegin
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
REQUEST_AST NoSuchFileFunction,
MD5_HASH *Md5HashPtr,
char *FileName,
char *ContentTypePtr
)
{
   int  status,
        FilePathLength;
   char  *cptr, *sptr, *zptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileBegin() !&F !&A !&A !16&H !&Z !&Z",
                 &FileBegin, NextTaskFunction, NoSuchFileFunction,
                 Md5HashPtr, FileName, ContentTypePtr);

   if (ERROR_REPORTED (rqptr))
   {
      /* previous error, cause threaded processing to unravel */
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /* if not previously initialized */
   if (!(tkptr = rqptr->FileTaskPtr) || !tkptr->TaskInitialized)
      tkptr = FileNewTask (rqptr);

   tkptr->ContentTypePtr = ContentTypePtr;
   tkptr->NextTaskFunction = NextTaskFunction;
   tkptr->NoSuchFileFunction = NoSuchFileFunction;

   if (rqptr->rqPathSet.AcceptLangChar &&
       ((ContentTypePtr && strsame (ContentTypePtr, "text/", 5)) ||
        rqptr->rqContentInfo.TypeText ||
        rqptr->rqPathSet.AcceptLangWildcard))
   {
      /* if a default language specified and default is satisfactory */
      if (!rqptr->rqPathSet.AcceptLangPtr || !FileAcceptLangDefault (rqptr))
      {
         /* try and resolve a language-specific document */
         if (!FileName)
         {
            /* must be a cache search, not interested at this stage */
            FileEnd (rqptr);
            return;
         }
         FileAcceptLangBegin (rqptr, FileName);
         return;
      }
   }

   if (!(cptr = FileName)) cptr = "";
   zptr = (sptr = tkptr->FileName) + sizeof(tkptr->FileName);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      FileEnd (rqptr);
      return;
   }
   *sptr = '\0';
   tkptr->FileNameLength = sptr - tkptr->FileName;

   if (Md5HashPtr)
   {
      /* buffer the supplied resource hash in the task structure */
      memcpy (&tkptr->Md5Hash, Md5HashPtr, sizeof(MD5_HASH));
   }
   else
   {
      /* generate a hash representing the file name being accessed */
      Md5Digest (tkptr->FileName, tkptr->FileNameLength, &tkptr->Md5Hash);
   }

   /* no file name supplied indicates only searching the cache */
   if (tkptr->AuthorizePath && tkptr->FileNameLength)
   {
      /***********************/
      /* check authorization */
      /***********************/

      cptr = MapVmsPath (tkptr->FileName, rqptr);
      if (!*cptr)
      {
         /* MAPURL errors are returned with a leading null (historical ;^) */
         ErrorGeneral (rqptr, cptr+1, FI_LI);
         FileEnd (rqptr);
         return;
      }
      Authorize (rqptr, cptr, -1, &FileAuthorizationAst);
      if (VMSnok (rqptr->rqAuth.FinalStatus))
      {
         /* if asynchronous authentication is underway then just wait for it */
         if (rqptr->rqAuth.FinalStatus == AUTH_PENDING) return;
         FileEnd (rqptr);
         return;
      }
   }

   /* not to-be-authorized, or authorized ... just carry on regardless! */
   FileAuthorizationAst (rqptr);
} 

/*****************************************************************************/
/*
There is a default language set against the path.  Compare the request header
"Accept-Language:" first language (if any) to the default language.  If the
same return true, otherwise false.
*/ 
 
BOOL FileAcceptLangDefault (REQUEST_STRUCT *rqptr)

{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileAcceptLangDefault()");

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "LANGUAGE default:!AZ accept:!AZ",
                 rqptr->rqPathSet.AcceptLangPtr,
                 rqptr->rqHeader.AcceptLangPtr ?
                    rqptr->rqHeader.AcceptLangPtr : "(none)");

   if (!(cptr = rqptr->rqPathSet.AcceptLangPtr)) return (true);
   if (!(sptr = rqptr->rqHeader.AcceptLangPtr)) return (true);
   if (!*sptr || *sptr == '*') return (true);
   while (*cptr && *sptr && *sptr != ',' && *sptr != ';' &&
          tolower(*cptr) == tolower(*sptr))
   {
      cptr++;
      sptr++;
   }
   if (!*cptr && (!*sptr || *sptr == ',' || *sptr == ';'))
      return (true);
   else
      return (false);
} 

/*****************************************************************************/
/*
This series if functions attempts to resolve a language-specific document based
on the file name originally supplied.  A seachable file specification is
contructed and used to find all possible language variants of the file
originally specified.
*/ 
 
FileAcceptLangBegin
(
REQUEST_STRUCT *rqptr,
char *FileName
)
{
   BOOL  LowerCaseHit;
   int  cnt, status;
   char  *cptr, *sptr, *zptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAcceptLangBegin() !&Z", FileName);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE, "LANGUAGE !AZ", FileName);

   tkptr = rqptr->FileTaskPtr;

   cptr = FileName;
   zptr = (sptr = tkptr->FileName) + sizeof(tkptr->FileName);
   /* copy the file name until the end of directory */
   while (*cptr && sptr < zptr)
   {
      if (*cptr == ']' && *(unsigned short*)cptr != '][') break;
      *sptr++ = *cptr++;
   }
   LowerCaseHit = false;

   /* copy the file name to the type delimiting period */
   while (*cptr && sptr < zptr)
   {
#ifdef ODS_EXTENDED
      if (OdsExtended)
      {
         if (*cptr == '^' && *(unsigned short*)cptr == '^.')
         {
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            continue;
         }
      }
      if (*cptr == '.') break;
      if (islower(*cptr)) LowerCaseHit = true;
      *sptr++ = *cptr++;
#else /* ODS_EXTENDED */
      if (*cptr == '.') break;
      *sptr++ = *cptr++;
#endif /* ODS_EXTENDED */
   }

#ifdef ODS_EXTENDED
   /* introduce an escaping '^' for the original type delimiting period */
   if (OdsExtended &&
       *cptr && rqptr->rqPathSet.AcceptLangChar == '.' &&
       sptr < zptr)
      *sptr++ = '^';
#endif /* ODS_EXTENDED */

   if (rqptr->rqPathSet.AcceptLangTypeVariant)
   {
      /*********************/
      /* variant file type */
      /*********************/

      /* if there is no existing period add one */
      if (!*cptr && sptr < zptr) *sptr++ = '.';

      /* copy file type to end of string */
      while (*cptr && sptr < zptr)
      {
#ifdef ODS_EXTENDED
         if (islower(*cptr)) LowerCaseHit = true;
#endif /* ODS_EXTENDED */
         *sptr++ = *cptr++;
      }

      /* note the point at which we begin to add wildcarded variants */
      tkptr->AcceptLangVariantPtr = sptr;

      /* append the variant delimiting character and a wildcard */
      if (sptr < zptr) *sptr++ = rqptr->rqPathSet.AcceptLangChar;
      if (sptr < zptr) *sptr++ = '*';
   }
   else
   {
      /*********************/
      /* variant file name */
      /*********************/

      /* note the point at which we begin to add wildcarded variants */
      tkptr->AcceptLangVariantPtr = sptr;

      /* insert the variant delimiting character, then wildcard(s :^) */
      if (sptr < zptr) *sptr++ = rqptr->rqPathSet.AcceptLangChar;
      /*
         Bit of a shonky here.
         I'm reserving space in the filename for the largest possible
         ISO language string which (as far as I know) is 5 characters
         (e.g. "fr-BE", "fr-CA").  This makes reusing the file buffer
         space in FileAcceptLangSelect() much easier (I'm getting
         lazier faster than I'm actually getting older).  The multiple
         wildcards used here (seem to) make no difference for RMS.
      */
      for (cnt = FILE_ACCEPT_LANG_VARIANT_MAX; cnt && sptr < zptr; cnt--)
         *sptr++ = '*';

      /* copy file type to end of string */
      while (*cptr && sptr < zptr)
      {
#ifdef ODS_EXTENDED
         if (islower(*cptr)) LowerCaseHit = true;
#endif /* ODS_EXTENDED */
         *sptr++ = *cptr++;
      }
   }

   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      FileEnd (rqptr);
      return;
   }
   *sptr = '\0';
   tkptr->FileNameLength = sptr - tkptr->FileName;

   tkptr->AcceptLangLowerCase = LowerCaseHit;

   OdsParse (&tkptr->FileOds,
             tkptr->FileName, tkptr->FileNameLength, NULL, 0,
             0, &FileAcceptLangParseAst, rqptr);
} 

/*****************************************************************************/
/*
AST called from FileAcceptLangBegin() when asynchronous parse completes.  Check
the status and if OK begin the search.
*/

FileAcceptLangParseAst (struct FAB *FabPtr)

{
   int  status;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   rqptr = FabPtr->fab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAcceptLangParseAst() !&F sts:!&S stv:!&S",
                 &FileAcceptLangParseAst, FabPtr->fab$l_sts, FabPtr->fab$l_stv);

   tkptr = rqptr->FileTaskPtr;

   if (VMSnok (status = tkptr->FileOds.Fab.fab$l_sts))
   {
      /*****************/
      /* error parsing */
      /*****************/

      /* ensure only the original file name appears in any error messages */
      tkptr->FileName[tkptr->FileNameLength] = '\0';

      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->FileOds.Nam_fnb & NAM$M_SEARCH_LIST) && status == RMS$_DNF)
         status = RMS$_FNF;

      if (tkptr->NoSuchFileFunction)
      {
         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "FILE !&S %!&M", status, status);
         rqptr->HomePageStatus = status;
         FileEnd (rqptr);
         return;
      }

      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      FileEnd (rqptr);
      return;
   }

   OdsSearch (&tkptr->FileOds, &FileAcceptLangSearchAst, rqptr);
}

/*****************************************************************************/
/*
AST called from FileAcceptLangParseAst() and then from
FileAcceptLangSearchAst() when asynchronous search completes.  Check the status
and if OK buffer the resolved language component in a comma-separated list. 
The continue the search.  Once no-more-files status occurs call
FileAcceptLangSelect() to assess the language variants (if any).
*/

FileAcceptLangSearchAst (struct FAB *FabPtr)

{
   int  status;
   char  *cptr, *sptr, *zptr;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   rqptr = FabPtr->fab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAcceptLangSearchAst() !&F sts:!&S stv:!&S",
                 &FileAcceptLangSearchAst, FabPtr->fab$l_sts,
                 FabPtr->fab$l_stv);

   tkptr = rqptr->FileTaskPtr;

   if (VMSnok (status = tkptr->FileOds.Fab.fab$l_sts))
   {
      if (status == RMS$_FNF || status == RMS$_NMF)
      {
         /***************************/
         /* end of directory search */
         /***************************/

         tkptr->FileOds.ParseInUse = false;
         FileAcceptLangSelect (rqptr);
         return;
      }

      /**********************/
      /* sys$search() error */
      /**********************/

      /* ensure only the original file name appears in any error messages */
      tkptr->FileName[tkptr->FileNameLength] = '\0';

      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      FileEnd (rqptr);
      return;
   }

   /****************/
   /* add language */
   /****************/

   if (!tkptr->AcceptLangTypesSize)
   {
      tkptr->AcceptLangTypesPtr = VmGetHeap (rqptr, FILE_ACCEPT_LANG_SIZE);
      tkptr->AcceptLangTypesSize = FILE_ACCEPT_LANG_SIZE;
   }
   zptr = (sptr = tkptr->AcceptLangTypesPtr) + tkptr->AcceptLangTypesSize;
   sptr += tkptr->AcceptLangTypesLength;
   if (tkptr->AcceptLangTypesLength && sptr < zptr) *sptr++ = ',';

   if (rqptr->rqPathSet.AcceptLangTypeVariant)
   {
      /* variant file type */
      cptr = tkptr->FileOds.NamVersionPtr;
      while (cptr > tkptr->FileOds.NamTypePtr &&
            *cptr != rqptr->rqPathSet.AcceptLangChar) cptr--;
      if (*cptr) cptr++;
      while (cptr < tkptr->FileOds.NamVersionPtr && sptr < zptr)
         *sptr++ = *cptr++;
   }
   else
   {
      /* variant file name */
      cptr = tkptr->FileOds.NamTypePtr;
      while (cptr > tkptr->FileOds.NamNamePtr &&
            *cptr != rqptr->rqPathSet.AcceptLangChar) cptr--;
      if (*cptr) cptr++;
      while (cptr < tkptr->FileOds.NamTypePtr && sptr < zptr)
         *sptr++ = *cptr++;
   }

   /* check the length of the (potential) variant (may not be) */
   if ((sptr - tkptr->AcceptLangTypesPtr) - tkptr->AcceptLangTypesLength <=
       FILE_ACCEPT_LANG_VARIANT_MAX)
   {
      if (sptr >= zptr)
      {
         /* more than FILE_ACCEPT_LANG_SIZE / _VARIANT (256/5~=51!!) */
         ErrorNoticed (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
         ErrorGeneral (rqptr, ErrorSanityCheck, FI_LI);
         FileEnd (rqptr);
         return;
      }
      *sptr = '\0';
      tkptr->AcceptLangTypesLength = sptr - tkptr->AcceptLangTypesPtr;
   }

   OdsSearch (&tkptr->FileOds, &FileAcceptLangSearchAst, rqptr);
}

/*****************************************************************************/
/*
Explicitly called by FileAcceptLangSearchAst() when no-more-files status occurs
to assess the comma-separated list of language components generated (if any).
By comparing each "Accept-Language:" entry against each of the file language
components select the first to match (if any).  If one matches then adjust the
original file name to include the language component with the type.  If none
matches revert to the original file name.  Commence file processing.
*/

FileAcceptLangSelect (REQUEST_STRUCT *rqptr)

{
   BOOL  AcceptLangVariant;
   int  status;
   char  *cptr, *sptr, *tptr, *zptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   tkptr = rqptr->FileTaskPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAcceptLangSelect() !&F !&Z !&Z",
                 &FileAcceptLangSelect, tkptr->AcceptLangTypesPtr,
                 rqptr->rqHeader.AcceptLangPtr);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "LANGUAGE variants:!AZ accept:!AZ",
                 tkptr->AcceptLangTypesPtr ?
                    tkptr->AcceptLangTypesPtr : "(none)",
                 rqptr->rqHeader.AcceptLangPtr ?
                    rqptr->rqHeader.AcceptLangPtr : "(none)");

   AcceptLangVariant = false;
   if (tkptr->AcceptLangTypesLength)
   {
      /*********************************/
      /* at least one possible variant */
      /*********************************/

      cptr = rqptr->rqHeader.AcceptLangPtr;
      while (*cptr)
      {
         while (*cptr && ISLWS(*cptr)) cptr++;
         if (!*cptr) break;
         sptr = tkptr->AcceptLangTypesPtr;
         while (*sptr)
         {
            tptr = cptr;
            while (*tptr && *tptr != ',' && *tptr != ';' &&
                   *sptr && *sptr != ',' &&
                   tolower(*tptr) == tolower(*sptr))
            {
               tptr++;
               sptr++;
            }
            if ((!*tptr || *tptr == ',' || *tptr == ';') &&
                (!*sptr || *sptr == ','))
            {
               AcceptLangVariant = true;
               break;
            }
            while (*sptr && *sptr != ',') sptr++;
            if (*sptr) sptr++;
         }
         if (AcceptLangVariant) break;
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) cptr++;
      }
   }

   if (AcceptLangVariant)
   {
      /**********************************/
      /* add language variant component */
      /**********************************/

      /*
         This handles both the post-file-type language variant, where the
         wildcard has just been appended to the file type, as well as the
         post-file-name language variant, where the wildcard was inserted
         between the end of the file name and the type delimiting period
         (in this case it reserved space using 5 wildcards).
      */
      zptr = tkptr->FileName + sizeof(tkptr->FileName);
      /* step over any period-escaping character (would be EFS only) */
      if (*(sptr = tkptr->AcceptLangVariantPtr) == '^') sptr++;
      /* step over the variant delimiting character */
      sptr++;
      /* 'cptr' has been left pointing at the matching language */
      tptr = cptr;
      if (tkptr->AcceptLangLowerCase)
         while (*cptr && *cptr != ',' && *cptr != ';' && sptr < zptr)
            *sptr++ = tolower(*cptr++);
      else
         while (*cptr && *cptr != ',' && *cptr != ';' && sptr < zptr)
            *sptr++ = toupper(*cptr++);
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         FileEnd (rqptr);
         return;
      }
      /* absorb any remaining wildcards */
      for (cptr = sptr; *cptr == '*'; cptr++);
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         FileEnd (rqptr);
         return;
      }
      *sptr = '\0';
      tkptr->FileNameLength = sptr - tkptr->FileName;

      if (rqptr->rqPathSet.AcceptLangTypeVariant)
      {
         /* (re)determine the (possibly) new content-type */
         while (sptr > tkptr->FileName && *sptr != '.') sptr--;
         ConfigContentType (&rqptr->rqContentInfo, sptr);
         tkptr->ContentTypePtr = rqptr->rqContentInfo.ContentTypePtr;
      }
   }
   else
   {
      /*******************************/
      /* no language variant, revert */
      /*******************************/

      /* eliminate any period-escaping character (would be EFS only) */
      if (*(sptr = cptr = tkptr->AcceptLangVariantPtr) == '^') cptr++;
      /* eliminate the variant delimiting character and the wildcard(s) */
      cptr++;
      while (*cptr && *cptr == '*') cptr++;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
      tkptr->FileNameLength = sptr - tkptr->FileName;
   }

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "LANGUAGE !AZ", tkptr->FileName);

   /*************************/
   /* begin processing file */
   /*************************/

   /* generate a hash representing the file name being accessed */
   Md5Digest (tkptr->FileName, tkptr->FileNameLength, &tkptr->Md5Hash);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                 "FILE !16&H !AZ", &tkptr->Md5Hash, tkptr->FileName);

   /* no file name supplied indicates only searching the cache */
   if (tkptr->AuthorizePath && tkptr->FileNameLength)
   {
      /***********************/
      /* check authorization */
      /***********************/

      cptr = MapVmsPath (tkptr->FileName, rqptr);
      if (!*cptr)
      {
         /* MAPURL errors are returned with a leading null (historical ;^) */
         ErrorGeneral (rqptr, cptr+1, FI_LI);
         FileEnd (rqptr);
         return;
      }
      Authorize (rqptr, cptr, -1, &FileAuthorizationAst);
      if (VMSnok (rqptr->rqAuth.FinalStatus))
      {
         /* if asynchronous authentication is underway then just wait for it */
         if (rqptr->rqAuth.FinalStatus == AUTH_PENDING) return;
         FileEnd (rqptr);
         return;
      }
   }

   /* not to-be-authorized, or authorized ... just carry on regardless! */
   FileAuthorizationAst (rqptr);
} 

/*****************************************************************************/
/*
This function provides an AST target is Authorize()ation ended up being done
asynchronously, otherwise it is just called directly to continue the modules
processing.
*/

FileAuthorizationAst (REQUEST_STRUCT *rqptr)

{
   int  status;
   char  *cptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAuthorizationAst() !&F !&S",
                 &FileAuthorizationAst, rqptr->rqAuth.FinalStatus);

   if (VMSnok (rqptr->rqAuth.FinalStatus))
   {
      FileEnd (rqptr);
      return;
   }

   tkptr = rqptr->FileTaskPtr;

   if (!tkptr->ContentTypePtr)
      tkptr->ContentTypePtr = ConfigContentTypeUnknown;
   if (ConfigSameContentType (tkptr->ContentTypePtr,
                              ConfigContentTypeUnknown, -1))
   {
      /* if the content-type is not known then use the default */
      if (Config.cfContent.ContentTypeDefaultPtr[0])
         tkptr->ContentTypePtr = Config.cfContent.ContentTypeDefaultPtr;
      else
         tkptr->ContentTypePtr = ConfigDefaultFileContentType;
   }

   if (!rqptr->rqCache.DoNotCache)
   {
      status = CacheSearch (rqptr);
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "CacheSearch() !&S", status);

      /* success status indicates the file is being supplied from cache */
      if (VMSok (status)) return;

      /* no file name supplied indicates only searching the cache */
      if (!tkptr->FileNameLength)
      {
         FileEnd (rqptr);
         return;
      }
   }

   if (rqptr->rqAuth.VmsUserProfileLength)
   {
      /*****************************************/
      /* check VMS-authenticated user's access */
      /*****************************************/

      status = AuthVmsCheckUserAccess (rqptr, tkptr->FileName,
                                       tkptr->FileNameLength);
      if (VMSnok (status))
      {
         if (tkptr->NoSuchFileFunction)
         {
            rqptr->HomePageStatus = status;
            FileEnd (rqptr);
            return;
         }
         if (status == SS$_NOPRIV)
         {
            rqptr->rqResponse.ErrorTextPtr =
               MapVmsPath (tkptr->FileName, rqptr);
            rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
            ErrorVmsStatus (rqptr, status, FI_LI);
         }
         FileEnd (rqptr);
         return;
      }
   }

   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (1, &SysPrvMask, 0, 0);
   OdsParse (&tkptr->FileOds, tkptr->FileName, tkptr->FileNameLength,
             ".", 1, 0, &FileParseAst, rqptr);
   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (0, &SysPrvMask, 0, 0);
} 

/*****************************************************************************/
/*
After checking the cache entry some problem prevented it's use (entry stale,
ACP access failed).  If it was a cache search only (no file name supplied) just
end processing.  If processing a specified file then restart with the parse (as
happens in FileAuthorizationAst()).
*/

FileCacheStale (REQUEST_STRUCT *rqptr)

{
   int  status;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileCacheStale()");

   tkptr = rqptr->FileTaskPtr;

   if (!tkptr->FileNameLength)
   {
      FileEnd (rqptr);
      return;
   }

   if (rqptr->rqAuth.VmsUserProfileLength)
   {
      /*****************************************/
      /* check VMS-authenticated user's access */
      /*****************************************/

      status = AuthVmsCheckUserAccess (rqptr, tkptr->FileName,
                                       tkptr->FileNameLength);
      if (VMSnok (status))
      {
         if (tkptr->NoSuchFileFunction)
         {
            rqptr->HomePageStatus = status;
            FileEnd (rqptr);
            return;
         }
         if (status == SS$_NOPRIV)
         {
            rqptr->rqResponse.ErrorTextPtr =
               MapVmsPath (tkptr->FileName, rqptr);
            rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
            ErrorVmsStatus (rqptr, status, FI_LI);
         }
         FileEnd (rqptr);
         return;
      }
   }

   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (1, &SysPrvMask, 0, 0);
   OdsParse (&tkptr->FileOds, tkptr->FileName, tkptr->FileNameLength,
             ".", 1, 0, &FileParseAst, rqptr);
   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (0, &SysPrvMask, 0, 0);
} 

/*****************************************************************************/
/*
AST called from FileAuthorizeAst() when asynchronous parse completes.  If
status OK set up and queue an ACP QIO to get file size and revision date/time,
ASTing to FileAcpInfoAst().
*/

FileParseAst (struct FAB *FabPtr)

{
   int  status;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   rqptr = FabPtr->fab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileParseAst() !&F sts:!&S stv:!&S",
                 &FileParseAst, FabPtr->fab$l_sts, FabPtr->fab$l_stv);

#if WATCH_MOD
   HttpdCheckPriv (FI_LI);
#endif /* WATCH_MOD */

   tkptr = rqptr->FileTaskPtr;

   if (VMSok (status = tkptr->FileOds.Fab.fab$l_sts))
   {
      if (strsame (tkptr->FileOds.NamTypePtr, ".DIR;", 5))
      {
         /* no sneaky getting directory contents by downloading files! */
         status = SS$_NOPRIV;
      }
      else
      if (rqptr->RemoteUser[0] &&
          strsame (tkptr->FileOds.NamTypePtr,
                    HTL_FILE_TYPE, sizeof(HTL_FILE_TYPE)-1) &&
          *(tkptr->FileOds.NamTypePtr+sizeof(HTL_FILE_TYPE)-1) == ';')
      {
         /* access to an HTL is OK *IF* it is authorized!! (for admin) */
         tkptr->ContentTypePtr = "text/plain";
      }
      else
      if ((strsame (tkptr->FileOds.NamTypePtr,
                    HTA_FILE_TYPE, sizeof(HTA_FILE_TYPE)-1) &&
           *(tkptr->FileOds.NamTypePtr+sizeof(HTA_FILE_TYPE)-1) == ';') ||
          (strsame (tkptr->FileOds.NamTypePtr,
                    HTL_FILE_TYPE, sizeof(HTL_FILE_TYPE)-1) &&
           *(tkptr->FileOds.NamTypePtr+sizeof(HTL_FILE_TYPE)-1) == ';'))
      {
         /* attempt to retrieve an HTA/HTL authorization file, scotch that! */
         status = SS$_NOSUCHFILE;
      }
   }

   /* if a wildcard the ACP function will return the first matching file! */
   if (tkptr->FileOds.Nam_fnb & NAM$M_WILDCARD) status = RMS$_WLD;

   if (VMSnok (status))
   {
      /*****************/
      /* error parsing */
      /*****************/

      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->FileOds.Nam_fnb & NAM$M_SEARCH_LIST) && status == RMS$_DNF)
         status = RMS$_FNF;

      if (tkptr->NoSuchFileFunction)
      {
         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "FILE !&S %!&M", status, status);
         rqptr->HomePageStatus = status;
         FileEnd (rqptr);
         return;
      }

      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      FileEnd (rqptr);
      return;
   }

   if (tkptr->FileOds.Nam_fnb & NAM$M_SEARCH_LIST &&
       !tkptr->SearchListCount++)
   {
      /*******************************/
      /* search to get actual device */
      /*******************************/

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "SEARCH-LIST");

      if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (1, &SysPrvMask, 0, 0);
      OdsSearchNoConceal (&tkptr->FileOds, &FileParseAst, rqptr);
      if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (0, &SysPrvMask, 0, 0);
      return;
   }

   /************/
   /* ACP info */
   /************/

   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (1, &SysPrvMask, 0, 0);
   OdsFileAcpInfo (&tkptr->FileOds, &FileAcpInfoAst, rqptr); 
   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (0, &SysPrvMask, 0, 0);
}

/****************************************************************************/
/*
AST called from FileParseAST() when ACP QIO completes.  If status indicates no
such file then call any file open error processing function originally
supplied, otherwise report the error.  If status OK call a function to open the
file, which function will AST to FileOpenAst().
*/

FileAcpInfoAst (REQUEST_STRUCT *rqptr)

#define FAT$C_UNDEFINED 0
#define FAT$C_FIXED 1
#define FAT$C_VARIABLE 2
#define FAT$C_VFC 3
#define FAT$C_STREAM 4
#define FAT$C_STMLF 5
#define FAT$C_STMCR 6

{
   int  status,
        ContentLength,
        Length;
   unsigned long  EndOfFileVbn;
   char  *cptr,
         *rfmptr,
         *KeepAlivePtr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileAcpInfoAst() !&F !&S", &FileAcpInfoAst,
                 rqptr->FileTaskPtr->FileOds.AcpInfo.AcpIOsb.Status);

#if WATCH_MOD
   HttpdCheckPriv (FI_LI);
#endif /* WATCH_MOD */

   tkptr = rqptr->FileTaskPtr;

   /* first, deassign the channel allocated by OdsFileAcpInfo() */
   sys$dassgn (tkptr->FileOds.AcpInfo.Channel);

   if ((status = tkptr->FileOds.AcpInfo.AcpIOsb.Status) == SS$_NOSUCHFILE)
      status = RMS$_FNF;

   if (status == RMS$_FNF && tkptr->NoSuchFileFunction)
   {
      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "FILE !&S %!&M", status, status);
      rqptr->HomePageStatus = status;
      FileEnd (rqptr);
      return;
   }
   else
   {
      /* the last point at which no-such-file could have been reported */
      tkptr->NoSuchFileFunction = NULL;

      if (VMSnok (status))
      {
         if (!rqptr->AccountingDone++)
            InstanceGblSecIncrLong (&AccountingPtr->DoFileCount);
         rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
         rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
         ErrorVmsStatus (rqptr, status, FI_LI);
         FileEnd (rqptr);
         return;
      }
   }

   if (rqptr->rqPathSet.CharsetPtr)
   {
      cptr = tkptr->FileOds.ResFileName;
      if (!*cptr) cptr = tkptr->FileOds.ExpFileName;
      rqptr->rqPathSet.CharsetPtr = FileSetCharset (rqptr, cptr);
   }

   if (tkptr->FileOds.AcpInfo.RecAttr.RecordType == FAT$C_VARIABLE ||
       tkptr->FileOds.AcpInfo.RecAttr.RecordType == FAT$C_VFC ||
       (tkptr->FileOds.AcpInfo.RecAttr.RecordType == FAT$C_FIXED &&
        (tkptr->FileOds.AcpInfo.RecAttr.RecordSize & 1)))
   {
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "record");
      tkptr->RecordIO = true;
   }

   /***************/
   /* file exists */
   /***************/

   EndOfFileVbn = tkptr->FileOds.AcpInfo.RecAttr.EndOfFileVbnLo +
                  (tkptr->FileOds.AcpInfo.RecAttr.EndOfFileVbnHi << 16);

   /* buffer these for possible use by the cache module */
   tkptr->EndOfFileVbn = EndOfFileVbn;
   tkptr->FirstFreeByte = tkptr->FileOds.AcpInfo.RecAttr.FirstFreeByte;

   if (EndOfFileVbn <= 1)
      tkptr->SizeInBytes = tkptr->FileOds.AcpInfo.RecAttr.FirstFreeByte;
   else
      tkptr->SizeInBytes = ((EndOfFileVbn-1) << 9) +
                           tkptr->FileOds.AcpInfo.RecAttr.FirstFreeByte;

   memcpy (&tkptr->ModifiedBinaryTime, &tkptr->FileOds.AcpInfo.RdtBinTime, 8);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
   {
      switch (tkptr->FileOds.AcpInfo.RecAttr.RecordType)
      {
         case FAT$C_VARIABLE :  rfmptr = "VAR"; break;
         case FAT$C_VFC :       rfmptr = "VFC"; break;
         case FAT$C_FIXED :     rfmptr = "FIX"; break;
         case FAT$C_STREAM :    rfmptr = "STM"; break;
         case FAT$C_STMLF :     rfmptr = "STMLF"; break;
         case FAT$C_STMCR :     rfmptr = "STMCR"; break;
         case FAT$C_UNDEFINED : rfmptr = "UDF"; break;
         default : rfmptr = "?";
      }
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
#ifdef ODS_EXTENDED
                 "FILE ODS:!UL rfm:!AZ ebk:!UL ffb:!UL (!UL bytes!AZ) rdt:!%D",
                 rqptr->PathOds,
#else
                 "FILE rfm:!AZ ebk:!UL ffb:!UL (!UL bytes!AZ) rdt:!%D",
#endif /* ODS_EXTENDED */
                 rfmptr, tkptr->EndOfFileVbn, tkptr->FirstFreeByte,
                 tkptr->SizeInBytes, tkptr->RecordIO ? " approx." : "",
                 &tkptr->ModifiedBinaryTime);
   }

   rqptr->AccountingDone++;
   InstanceGblSecIncrLong (&AccountingPtr->DoFileCount);

   if (!rqptr->rqResponse.HeaderPtr &&
       !tkptr->ContentHandlerFunction)
   {
      /**************************/
      /* full response required */
      /**************************/

      /*
         For variable-length, etc. files (accessed using record-mode) file size
         cannot be accurately determined.  Indicate this using length of -1.
      */
      if (tkptr->RecordIO)
         status = FileResponseHeader (rqptr, tkptr->ContentTypePtr,
                                      -1, &tkptr->ModifiedBinaryTime);
      else
         status = FileResponseHeader (rqptr, tkptr->ContentTypePtr,
                                      tkptr->SizeInBytes,
                                      &tkptr->ModifiedBinaryTime);
      if (VMSnok (status))
      {
         if (status == LIB$_NEGTIM)
            InstanceGblSecIncrLong (&AccountingPtr->DoFileNotModifiedCount);
         FileEnd (rqptr);
         return;
      }

      /* quit here if the HTTP method was HEAD */
      if (rqptr->rqHeader.Method == HTTP_METHOD_HEAD)
      {
         FileEnd (rqptr);
         return;
      }
   }

   /*****************/
   /* NOW open file */
   /*****************/

   /* "temporary" file, automatic delete on closing it */
   if (rqptr->DeleteOnClose)
   {
      tkptr->FileOds.Fab.fab$l_fop |= FAB$M_DLT;
      tkptr->FileOds.DeleteOnClose = true;
   }

   /* initialize the date, header and protection extended attribute blocks */
   tkptr->FileOds.Fab.fab$l_xab = &tkptr->FileOds.XabDat;
   tkptr->FileOds.XabDat = cc$rms_xabdat;
   tkptr->FileOds.XabDat.xab$l_nxt = &tkptr->FileOds.XabFhc;
   tkptr->FileOds.XabFhc = cc$rms_xabfhc;
   tkptr->FileOds.XabFhc.xab$l_nxt = &tkptr->FileOds.XabPro;
   tkptr->FileOds.XabPro = cc$rms_xabpro;

   tkptr->FileOds.Fab.fab$l_ctx = rqptr;

   FileOpen (tkptr);
} 

/*****************************************************************************/
/*
Match the supplied file name (generally needs to be a NOCONCEAL sys$search()
result file name as provided by OdsSearchNoConceal() with the 'name=value' pair
'name' string of a "SET /path/* charset=(pattern,charset)" rule.  If a
successful match return the corresponding 'value' otherwise return NULL.
*/

char* FileSetCharset
(
REQUEST_STRUCT *rqptr,
char *FileName
)
{
   static char  Charset [64];

   int  status;
   char  *cptr;
   char  FileSpec [128];

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileSetCharset() !&Z !&Z",
                 rqptr->rqPathSet.CharsetPtr, FileName);

   if (!(cptr = rqptr->rqPathSet.CharsetPtr)) return (NULL);
   if (cptr[0] != '(') return (cptr);
   if (!FileName) return (NULL);

   for (;;)
   {
      status = StringParseNameValue (&cptr, false,
                                     FileSpec, sizeof(FileSpec),
                                     Charset, sizeof(Charset));
      if (VMSnok (status)) return (NULL);
      if (StringMatch (NULL, FileName, FileSpec)) return (Charset);
   }
}

/****************************************************************************/
/*
Generate a response header suitable for the file being returned.
This may be a 304 (not modified) if appropriate.
This function also used by the CACHE.C module.
*/

FileResponseHeader
(
REQUEST_STRUCT *rqptr,
char *ContentType,
int ContentLength,
unsigned long *RdtBinTimePtr
)
{
   int  status;
   char  *ContentTypePtr,
         *KeepAlivePtr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileResponseHeader()");

   tkptr = rqptr->FileTaskPtr;

   if ((rqptr->rqHeader.Method == HTTP_METHOD_GET ||
        rqptr->rqHeader.Method == HTTP_METHOD_HEAD) &&
       (rqptr->rqTime.IfModifiedSinceVMS64bit[0] ||
        rqptr->rqTime.IfModifiedSinceVMS64bit[1]) &&
        !rqptr->PragmaNoCache &&
        !rqptr->DeleteOnClose)
   {
      /*********************/
      /* if modified since */
      /*********************/

      status = HttpIfModifiedSince (rqptr, RdtBinTimePtr, ContentLength);
      if (VMSnok (status))
      {
         /* returned status is LIB$_NEGTIM if not modified/sent */
         return (status);
      }
   }

   if (!rqptr->AccountingDone++)
      InstanceGblSecIncrLong (&AccountingPtr->DoFileCount);

   if (Config.cfTimeout.KeepAlive &&
       rqptr->KeepAliveRequest &&
       ContentLength > 0)
   {
      rqptr->KeepAliveResponse = true;
      rqptr->rqNet.KeepAliveCount++;
      KeepAlivePtr = DEFAULT_KEEPALIVE_HTTP_HEADER;
   }
   else
      KeepAlivePtr = "";

   if (rqptr->rqHeader.QueryStringPtr[0] &&
       tolower(rqptr->rqHeader.QueryStringPtr[0]) == 'h' &&
       strsame (rqptr->rqHeader.QueryStringPtr, "httpd=content", 13))
   {
      /* request-specified content-type (default to plain-text) */
      if (strsame (rqptr->rqHeader.QueryStringPtr+13, "&type=", 6))
         ContentTypePtr = rqptr->rqHeader.QueryStringPtr + 19;
      else
         ContentTypePtr = "text/plain";
   }
   else
      ContentTypePtr = ContentType;

   ResponseHeader (rqptr, 200, ContentTypePtr,
                   ContentLength, RdtBinTimePtr, KeepAlivePtr);

   return (SS$_NORMAL);
} 

/*****************************************************************************/
/*
Complete the FAB according to record or block mode transfer.  Asynchronous
sys$open() ASTing to FileOpenAst().  FileOpenAst() can call this function
directly again, see note on locked files.
*/

FileOpen (FILE_TASK *tkptr)

{

   int  status;
   REQUEST_STRUCT  *rqptr;

   /*********/
   /* begin */
   /*********/

   rqptr = tkptr->FileOds.Fab.fab$l_ctx;

   if (WATCH_MOD)
   {
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileOpen() !&F", &FileOpen);
   }

   /* get the 'fid' and use the NAM block to identify the file */
#ifdef ODS_EXTENDED
   if (OdsExtended)
   {
      memcpy (&tkptr->FileOds.Naml.naml$w_fid,
              tkptr->FileOds.AcpInfo.FileFib.fib$w_fid,
              sizeof(tkptr->FileOds.AcpInfo.FileFib.fib$w_fid));
   }
   else
#endif /* ODS_EXTENDED */
   {
      memcpy (&tkptr->FileOds.Nam.nam$w_fid,
              tkptr->FileOds.AcpInfo.FileFib.fib$w_fid,
              sizeof(tkptr->FileOds.AcpInfo.FileFib.fib$w_fid));
   }
   tkptr->FileOds.Fab.fab$l_fop |= FAB$M_NAM;

   if (tkptr->RecordIO)
   {
      tkptr->FileOds.Fab.fab$b_fac = FAB$M_GET;
      tkptr->FileOds.Fab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;
   }
   else
   {
      tkptr->FileOds.Fab.fab$b_fac = FAB$M_GET | FAB$M_BIO; 
      tkptr->FileOds.Fab.fab$b_shr = FAB$M_SHRGET;
   }

   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (1, &SysPrvMask, 0, 0);
   tkptr->FileOds.Fab.fab$l_fop |= FAB$M_ASY;
   sys$open (&tkptr->FileOds.Fab, &FileOpenAst, &FileOpenAst);
   if (rqptr->rqAuth.VmsUserProfileLength) sys$setprv (0, &SysPrvMask, 0, 0);
} 

/*****************************************************************************/
/*
AST called from FileOpen() when asynchronous open completes.  Initiate an
asynchronous RAB connect.
*/

FileOpenAst (struct FAB *FabPtr)

{
   int  status,
        MultiBlockCount;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   rqptr = FabPtr->fab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileOpenAst() !&F sts:!&S stv:!&S",
                 &FileOpenAst, FabPtr->fab$l_sts, FabPtr->fab$l_stv);

#if WATCH_MOD
   HttpdCheckPriv (FI_LI);
#endif /* WATCH_MOD */

   tkptr = rqptr->FileTaskPtr;

   status = tkptr->FileOds.Fab.fab$l_sts;
   if (status == RMS$_FLK && !tkptr->RecordIO)
   {
      /* RMS$_FLK, block-mode cannot SHRPUT, try record-mode! */
      tkptr->RecordIO = true;
      FileOpen (tkptr);
      return;
   }
   if (VMSnok (status))
   {
      /* ensure the response header created above is not used for the error */
      rqptr->rqResponse.HeaderPtr = NULL;
      rqptr->rqResponse.HeaderLength = 0;
      rqptr->rqResponse.HttpStatus = 0;

      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->FileOds.Nam_fnb & NAM$M_SEARCH_LIST) &&
          status == RMS$_DNF)
         status = RMS$_FNF;

      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      FileEnd (rqptr);
      return;
   }

   tkptr->FileOds.Rab = cc$rms_rab;
   tkptr->FileOds.Rab.rab$l_ctx = rqptr;
   tkptr->FileOds.Rab.rab$l_fab = &tkptr->FileOds.Fab;
   if (tkptr->RecordIO)
   {
      /* divide by 512 and one for trailing bytes */
      MultiBlockCount = (tkptr->SizeInBytes >> 9) + 1;
      if (MultiBlockCount > 127)
      {
         tkptr->FileOds.Rab.rab$b_mbc = 127;
         tkptr->FileOds.Rab.rab$b_mbf = 2;
      }
      else
      {
         /* less than 127 * 512 bytes in length, one buffer enough! */
         tkptr->FileOds.Rab.rab$b_mbc = MultiBlockCount;
         tkptr->FileOds.Rab.rab$b_mbf = 1;
      }
      tkptr->FileOds.Rab.rab$l_rop = RAB$M_RAH | RAB$M_ASY;

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "mbc:!UL mbf:!UL",
                    tkptr->FileOds.Rab.rab$b_mbc, tkptr->FileOds.Rab.rab$b_mbf);
   }
   else
   {
      tkptr->FileOds.Rab.rab$l_bkt = 0;
      tkptr->FileOds.Rab.rab$l_rop = RAB$M_RAH | RAB$M_BIO | RAB$M_ASY;
   }

   sys$connect (&tkptr->FileOds.Rab, &FileConnectAst, &FileConnectAst);
}

/*****************************************************************************/
/*
AST called from FileOpenAst() when asynchronous RAB connect completes.
Initiate stream-LF file conversion if the file meets requirements.  If
suitable for caching then initiate a cache load and begin reading the file
contents.
*/ 

FileConnectAst (struct RAB *RabPtr)

{
   int  status,
        BufferSize;
   char  *cptr, *sptr, *zptr;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;
   FILE_CONTENT  *fcptr;

   /*********/
   /* begin */
   /*********/

   rqptr = RabPtr->rab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileConnectAst() !&F sts:!&S stv:!&S",
                 &FileConnectAst, RabPtr->rab$l_sts, RabPtr->rab$l_stv);

   tkptr = rqptr->FileTaskPtr;

   if (VMSnok (status = RabPtr->rab$l_sts))
   {
      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      FileEnd (rqptr);
      return;
   }

   /******************************/
   /* stream-LF file conversion? */
   /******************************/

   if (Config.cfMisc.StreamLfConversionMaxKbytes &&
       tkptr->RecordIO &&
       tkptr->FileOds.Fab.fab$b_rfm == FAB$C_VAR &&
       rqptr->rqPathSet.StmLF &&
       ConfigSameContentType (tkptr->ContentTypePtr, "text/", 5) &&
       !ConfigSameContentType (tkptr->ContentTypePtr, ConfigContentTypeSsi, -1))
   {
      /* divide by two to get the number of kilobytes (1024) in the file */
      if ((tkptr->SizeInBytes >> 10) <= Config.cfMisc.StreamLfConversionMaxKbytes)
         StmLfBegin (MapVmsPath(tkptr->FileName, rqptr),
                     tkptr->FileName,
                     tkptr->FileNameLength,
                     rqptr->PathOdsExtended,
                     rqptr->rqAuth.VmsUserProfileLength);
   }

   /********************/
   /* begin processing */
   /********************/

   /*
      Cache load is not initiated if this is not a stand-alone file request
      (i.e. not part of some other activity, e.g. a directory read-me file,
      if it is already being loaded via another request, if a "temporary"
      file ('DeleteOnClose'), or via requests with a VMS authentication
      profile attached.
   */
   if (CacheEnabled &&
       tkptr->CacheAllowed &&
       !rqptr->rqCache.DoNotCache &&
       !rqptr->rqHeader.QueryStringPtr[0] &&
       !tkptr->EscapeHtml &&
       !rqptr->rqCache.NotUsable &&
       !rqptr->DeleteOnClose &&
       !rqptr->rqAuth.VmsUserProfileLength)
   {
      /* get a cache structure allocated */
      CacheLoadBegin (rqptr);
      /* 'rqptr->rqCache.EntryPtr' will now control cache load behaviour */
   }
   else
      rqptr->rqCache.EntryPtr = NULL;

   if (tkptr->ContentHandlerFunction)
   {
      /************************/
      /* buffer file contents */
      /************************/

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "!UL !UL",
                    tkptr->SizeInBytes, tkptr->FileContentsSizeMax);

      if (tkptr->SizeInBytes > tkptr->FileContentsSizeMax)
      {
         rqptr->rqResponse.HttpStatus = 500;
         ErrorGeneral (rqptr, ErrorBufferFileMaxBytes, FI_LI);
         FileEnd (rqptr);
         return;
      }
      if (!rqptr->rqCache.EntryPtr)
      {
         /* make buffer size a multiple of 512 byte blocks (for block I/O) */
         BufferSize = (tkptr->SizeInBytes / 512);
         if (tkptr->SizeInBytes % 512) BufferSize++;
         BufferSize *= 512;
      }
      else
      {
         /* size guarantees written to cache can be written to contents */
         BufferSize = rqptr->rqCache.EntryPtr->CacheSize;
      }

      /* allocate a buffer */
      rqptr->FileContentPtr = fcptr = (FILE_CONTENT*)
         VmGetHeap (rqptr, sizeof(FILE_CONTENT) + BufferSize);
      fcptr->ContentSize = BufferSize;
      /* buffer space immediately follows the structured storage */
      fcptr->ContentPtr = (char*)fcptr + sizeof(FILE_CONTENT);

      /* populate the file contents structure with some file data */
      zptr = (sptr = fcptr->FileName) + sizeof(fcptr->FileName);
      for (cptr = tkptr->FileName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         FileEnd (rqptr);
         return;
      }
      *sptr = '\0';
      fcptr->FileNameLength = sptr - fcptr->FileName;

      memcpy (&fcptr->CdtBinTime, &tkptr->FileOds.XabDat.xab$q_cdt, 8);
      memcpy (&fcptr->RdtBinTime, &tkptr->FileOds.XabDat.xab$q_rdt, 8);

      fcptr->UicGroup = tkptr->FileOds.XabPro.xab$w_grp;
      fcptr->UicMember = tkptr->FileOds.XabPro.xab$w_mbm;
      fcptr->Protection = tkptr->FileOds.XabPro.xab$w_pro;

      /* set the content structure handler to the supplied function */
      rqptr->FileContentPtr->ContentHandlerFunction = 
         tkptr->ContentHandlerFunction;

      /* none of these little tricks, just get the raw file! */
      tkptr->PreTagFileContents = tkptr->EscapeHtml = false;
   }
   else
   {
      /* network writes are checked for success, fudge the first one! */
      rqptr->rqNet.WriteIOsb.Status = SS$_NORMAL;
   }

   if (tkptr->PreTagFileContents)
   {
      tkptr->PreTagEndFileContents = true;
      NetWriteBuffered (rqptr, &FileEmptyBuffer, "<PRE>", 5);
      return;
   }

   FileEmptyBuffer (rqptr);
}

/*****************************************************************************/
/*
End of file transfer, successful or otherwise.
*/

FileEnd (REQUEST_STRUCT *rqptr)

{
   int  status;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileEnd() !&F", &FileEnd);

   tkptr = rqptr->FileTaskPtr;

   /* ensure the FAB user context is set for FileEnd2() */
   tkptr->FileOds.Fab.fab$l_ctx = rqptr;

   /* if the file was begin cached at the same time */
   if (rqptr->rqCache.EntryPtr) CacheLoadEnd (rqptr);

   if (tkptr->FileOds.Fab.fab$w_ifi)
      OdsClose (&tkptr->FileOds, &FileEnd2, rqptr);
   else
      FileEnd2 (&tkptr->FileOds.Fab);
} 

/*****************************************************************************/
/*
May be called directly or as an AST from OdsClose() above.  Ihe file has been
sys$close()d at this stage.  Indirectly or directly call the next task.
*/

FileEnd2 (struct FAB *FabPtr)

{
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   rqptr = FabPtr->fab$l_ctx;
   tkptr = rqptr->FileTaskPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileEnd2() !&F sts:!&S stv:!&S !&X !&X",
                 &FileEnd2, FabPtr->fab$l_sts, FabPtr->fab$l_stv,
                 tkptr->NoSuchFileFunction, rqptr->FileContentPtr);

#if WATCH_MOD
   HttpdCheckPriv (FI_LI);
#endif /* WATCH_MOD */

   /* release internal RMS parse structures */
   if (tkptr->FileOds.ParseInUse) OdsParseRelease (&tkptr->FileOds);

   /* reset this flag to indicate we've finished with the structure */
   tkptr->TaskInitialized = false;

   if (tkptr->NoSuchFileFunction)
   {
      /* file could not have been found, declare the appropriate handler */
      SysDclAst (tkptr->NoSuchFileFunction, rqptr);
      return;
   }

   if (rqptr->FileContentPtr)
   {
      /* next task gets control once the file has been content-handled */
      rqptr->FileContentPtr->NextTaskFunction = tkptr->NextTaskFunction;

      /* file contents loaded, now process using the specified handler */
      SysDclAst (rqptr->FileContentPtr->ContentHandlerFunction, rqptr);
      return;
   }

   if (rqptr->rqHeader.Method != HTTP_METHOD_HEAD &&
       tkptr->PreTagEndFileContents)
   {
      /* success, encapsulating a file, add the end tag */
      NetWriteBuffered (rqptr, tkptr->NextTaskFunction, "</PRE>\n", 7);
      return;
   }

   /* success, declare the next task */
   SysDclAst (tkptr->NextTaskFunction, rqptr);
} 

/*****************************************************************************/
/*
Ensure any output buffer contents are flushed.
*/

FileEmptyBuffer (REQUEST_STRUCT *rqptr)

{
   int  status;
   REQUEST_AST  AstFunction;
   FILE_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileEmptyBuffer() !&F !UL",
                 &FileEmptyBuffer, rqptr->rqOutput.BufferCount);

   tkptr = rqptr->FileTaskPtr;

   if (tkptr->RecordIO)
      AstFunction = &FileNextRecord;
   else
      AstFunction = &FileNextBlocks;
   if (rqptr->rqOutput.BufferCount)
   {
      /* need exclusive use, flush the current contents */
      NetWriteFullFlush (rqptr, AstFunction);
   }
   else
      SysDclAst (AstFunction, rqptr);
}

/*****************************************************************************/
/*
Queue a read of the next record from the file.  When the read completes call 
FileNextRecordAst() function to post-process the read and/or send the data 
to the client.  Don't bother to test any status here, the AST routine will do 
that!
*/ 

FileNextRecord (REQUEST_STRUCT *rqptr)

{
   FILE_TASK  *tkptr;
   FILE_CONTENT  *fcptr;
   REQUEST_CACHE  *rcaptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileNextRecord() !&F !&S",
                 &FileNextRecord, rqptr->rqNet.WriteIOsb.Status);

   tkptr = rqptr->FileTaskPtr;
   fcptr = rqptr->FileContentPtr;
   rcaptr = &rqptr->rqCache;

   if (!fcptr && VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      /* network write has failed (as AST), bail out now */
      FileEnd (rqptr);
      return;
   }

   if (!tkptr->FileOds.Rab.rab$l_ubf)
   {
      /* first I/O in file transfer, organise the buffer */
      if (rcaptr->EntryPtr)
      {
         /* using a cache buffer */
         rcaptr->ContentPtr =
            rcaptr->CurrentPtr = 
            rcaptr->EntryPtr->ContentPtr;
         rcaptr->ContentLength = 0;
         rcaptr->ContentRemaining = rcaptr->EntryPtr->CacheSize;
      }

      if (fcptr)
      {
         /* (perhaps also) using a file content buffer */
         fcptr->CurrentPtr = fcptr->ContentPtr;
         fcptr->ContentRemaining = fcptr->ContentSize;
         fcptr->ContentLength = 0;
      }

      if (!rcaptr->EntryPtr && !fcptr)
      {
         /* not loading cache or contents, initialize network buffer */
         NetWriteInit (rqptr);
      }
   }

   if (rcaptr->EntryPtr)
   {
      /* reading directly into file cache buffer */
      tkptr->FileOds.Rab.rab$l_ubf = rcaptr->CurrentPtr;
      if (rcaptr->ContentRemaining > 32767)
         tkptr->FileOds.Rab.rab$w_usz = 32767;
      else
         tkptr->FileOds.Rab.rab$w_usz = rcaptr->ContentRemaining;
   }
   else
   if (fcptr)
   {
      /* reading directly into file contents buffer */
      tkptr->FileOds.Rab.rab$l_ubf = fcptr->CurrentPtr;
      if (fcptr->ContentRemaining > 32767)
         tkptr->FileOds.Rab.rab$w_usz = 32767;
      else
         tkptr->FileOds.Rab.rab$w_usz = fcptr->ContentRemaining;
   }
   else
   {
      /* not loading cache or contents */
      tkptr->FileOds.Rab.rab$l_ubf = rqptr->rqOutput.BufferCurrentPtr;
      if (rqptr->rqOutput.BufferRemaining > 32767)
         tkptr->FileOds.Rab.rab$w_usz = 32767;
      else
         tkptr->FileOds.Rab.rab$w_usz = rqptr->rqOutput.BufferRemaining;
   }

   /* GET the next record (or last read if RFA is set) */
   sys$get (&tkptr->FileOds.Rab, &FileNextRecordAst, &FileNextRecordAst);
}

/*****************************************************************************/
/*
The read of the next record from the file has completed.  Post-process and/or 
queue a network write to the client.  When the network write completes it will 
call the function FileNextRecord() to queue a read of the next record. Record-
by-record processing builds up as many file records as possible into available 
buffer space before transfering them to the network as one I/O.  The 
sys$get()s use the buffer area pointed to by
'rqptr->rqOutput.BufferCurrentPtr' and the size specified by
'rqptr->rqOutput.BufferRemaining'.  If/when a sys$get() fails with a "record 
too big for user's buffer", because we're running out of buffer space, the 
accumulated records are flushed to the network and the pointer and available 
space values are reset.  The FileNextRecord() AST called when the 
NetWrite() completes will re-read a record that failed with the "record too 
big" error by forcing the (re)read to be by RFA for that one record.
*/ 

FileNextRecordAst (struct RAB *RabPtr)

{
   int  rsz, status,
        CacheBufferRemaining;
   char  *ubf;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;
   REQUEST_CACHE  *rcaptr;
   FILE_CONTENT  *fcptr;

   /*********/
   /* begin */
   /*********/

   rqptr = RabPtr->rab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
   {
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
"FileNextRecordAst() !&F sts:%X!XL stv:%X!XL rac:!2XL usz:!UL rsz:!UL",
                 &FileNextRecordAst, RabPtr->rab$l_sts, RabPtr->rab$l_stv,
                 RabPtr->rab$b_rac, RabPtr->rab$w_usz, RabPtr->rab$w_rsz);
      if (VMSok (RabPtr->rab$l_sts))
         WatchDataDump (RabPtr->rab$l_ubf, RabPtr->rab$w_rsz);
   }

   tkptr = rqptr->FileTaskPtr;
   fcptr = rqptr->FileContentPtr;
   rcaptr = &rqptr->rqCache;

   if (VMSnok (tkptr->FileOds.Rab.rab$l_sts))
   {
      /*
         Test for "record too big for user's buffer", indicates that it's
         time to flush the records accumulated in the buffer. This should
         never happen when reading into a cache or file contents buffer!!!
      */
      if (tkptr->FileOds.Rab.rab$l_sts == RMS$_RTB)
      {
         if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
            WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "RMS$_RTB");
         if (!rcaptr->EntryPtr && !fcptr)
         {
            /* not reading into cache or file contents buffer */
            if (rqptr->rqOutput.BufferCount == 0)
            {
               /* legitimate "record too big for user's buffer" error */
               rqptr->rqResponse.ErrorTextPtr =
                  MapVmsPath (tkptr->FileName, rqptr);
               rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
               ErrorVmsStatus (rqptr, RMS$_RTB, FI_LI);
               FileEnd (rqptr);
               return;
            }

            /* insufficient space for next record, flush buffer contents */
            FileFlushBuffer (rqptr, &FileNextRecord);

            /* re-read the record that just failed by forcing a read by RFA */
            tkptr->FileOds.Rab.rab$b_rac |= RAB$C_RFA;

            return;
         }
         else
         {
            /* THIS IS NOT SUPPOSED TO HAPPEN!!! (with cache/content buffer) */
            rqptr->rqResponse.ErrorTextPtr =
               MapVmsPath (tkptr->FileName, rqptr);
            rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
            ErrorVmsStatus (rqptr, RMS$_RTB, FI_LI);
            FileEnd (rqptr);
            return;
         }
      }

      if (tkptr->FileOds.Rab.rab$l_sts == RMS$_EOF)
      {
         if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
            WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "RMS$_EOF");
         if (rcaptr->EntryPtr)
         {
            /* a cache load becomes valid only after we get to end-of-file */
            rcaptr->EntryPtr->CacheLoadValid = true;
         }
         /* only need to flush if we're not buffering the file contents */
         if (!fcptr)
            FileFlushBuffer (rqptr, &FileEnd);
         else
            FileEnd (rqptr);
         return;
      }

      /* hmmm, file read error */
      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, tkptr->FileOds.Rab.rab$l_sts, FI_LI);
      FileEnd (rqptr);
      return;
   }

   /* force the next read to be sequential (in case the last was by RFA) */
   tkptr->FileOds.Rab.rab$b_rac &= ~RAB$C_RFA;

   rsz = tkptr->FileOds.Rab.rab$w_rsz;
   ubf = tkptr->FileOds.Rab.rab$l_ubf;
   if (rsz)
   {
      /* add newline if last character in line is not a newline! */
      if (ubf[rsz-1] != '\n') ubf[rsz++] = '\n';
   }
   else
   {
      /* must be a blank line (empty record), add a newline */
      ubf[rsz++] = '\n';
   }

   if (rcaptr->EntryPtr)
   {
      /* this section of the cache buffer has not filled or overflowed */
      rcaptr->ContentRemaining -= rsz;
      rcaptr->EntryPtr->ContentLength = (rcaptr->ContentLength += rsz);
      rcaptr->CurrentPtr += rsz;

      if (fcptr)
      {
         /* also copy into the contents buffer (always enough space) */
         memcpy (fcptr->CurrentPtr, ubf, rsz);
         fcptr->ContentRemaining -= rsz;
         fcptr->ContentLength += rsz;
         fcptr->CurrentPtr += rsz;
         FileNextRecord (rqptr);
         return;
      }

      if (rcaptr->CurrentPtr >= rcaptr->ContentPtr + OutputBufferSize)
      {
         /* not file contents, flush now we have arrived at an output chunk */
         FileFlushBuffer (rqptr, &FileNextRecord);
      }
      else
         FileNextRecord (rqptr);
      return;
   }

   if (fcptr)
   {
      /* just buffering the file contents, adjust the buffer read values */
      fcptr->ContentRemaining -= rsz;
      fcptr->ContentLength += rsz;
      fcptr->CurrentPtr += rsz;
      FileNextRecord (rqptr);
      return;
   }

   if (rqptr->rqOutput.BufferRemaining > rsz)
   {
      /* has not filled or overflowed the standard output buffer */
      rqptr->rqOutput.BufferRemaining -= rsz;
      rqptr->rqOutput.BufferCount += rsz;
      rqptr->rqOutput.BufferCurrentPtr += rsz;
      FileNextRecord (rqptr);
      return;
   }

   /* insufficient space for next record, flush buffer contents */
   FileFlushBuffer (rqptr, &FileNextRecord);
   /* re-read the record that just failed by forcing a read by RFA */
   tkptr->FileOds.Rab.rab$b_rac |= RAB$C_RFA;
}

/*****************************************************************************/
/*
Flush the file records accumulated in the 'Buffer'.  If plain text and 
encapsulated then escape any HTML-forbidden characters using a separate 
function.  This function is never called when buffering the contents of a file.
*/

FileFlushBuffer
(
REQUEST_STRUCT *rqptr,
REQUEST_AST AstFunction
)
{
   int  status;
   REQUEST_CACHE  *rcaptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileFlushBuffer() !&A !UL",
                 AstFunction, rqptr->rqOutput.BufferCount);

   rcaptr = &rqptr->rqCache;

   if (rqptr->FileTaskPtr->EscapeHtml)
      return (FileWriteBufferEscapeHtml (rqptr, AstFunction));
   else
   {
      if (rcaptr->EntryPtr)
      {
         /* flushing from a cache buffer */
         if (rcaptr->CurrentPtr == rcaptr->ContentPtr)
         {
            /* nothing to flush! */
            SysDclAst (AstFunction, rqptr);
            return;
         }

         NetWrite (rqptr, AstFunction,
                   rcaptr->ContentPtr,
                   rcaptr->CurrentPtr - rcaptr->ContentPtr);

         /* reset the pointer to the new flushed-from position */
         rcaptr->ContentPtr = rcaptr->CurrentPtr;
         return;
      }
      else
      {
         /* flushing from a standard output buffer */
         if (!rqptr->rqOutput.BufferCount)
         {
            /* nothing to flush! */
            SysDclAst (AstFunction, rqptr);
            return;
         }

         NetWrite (rqptr, AstFunction,
                   rqptr->rqOutput.BufferPtr,
                   rqptr->rqOutput.BufferCount);

         /* reset standard output buffer */
         NetWriteInit (rqptr);
         return;
      }
   }
}

/*****************************************************************************/
/*
Queue a read of the next series of Virtual Blocks from the file.  When the 
read completes call FileNextBlocksAst() function to post-process the read 
and/or send the data to the client.  Don't bother to test any status here, the 
AST routine will do that!
*/ 

FileNextBlocks (REQUEST_STRUCT *rqptr)

{
   unsigned int  BufferSize;
   FILE_TASK  *tkptr;
   FILE_CONTENT  *fcptr;
   REQUEST_CACHE  *rcaptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "FileNextBlocks() !&F !&S",
                 &FileNextBlocks, rqptr->rqNet.WriteIOsb.Status);

   tkptr = rqptr->FileTaskPtr;
   fcptr = rqptr->FileContentPtr;
   rcaptr = &rqptr->rqCache;

   if (!fcptr && VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      /* network write has failed (delivered via AST), bail out now */
      FileEnd (rqptr);
      return;
   }

   if (tkptr->EndOfBlockIO)
   {
      /* effective EOF for block I/O */
      tkptr->FileOds.Rab.rab$l_sts = RMS$_EOF;
      SysDclAst (FileNextBlocksAst, &tkptr->FileOds.Rab);
      return;
   }

   if (!tkptr->FileOds.Rab.rab$l_bkt)
   {
      /* first read, initialize read buffer space */
      tkptr->FileOds.Rab.rab$l_bkt = 1;

      if (rcaptr->EntryPtr)
      {
         /* populating a cache buffer */
         rcaptr->CurrentPtr = rcaptr->ContentPtr = rcaptr->EntryPtr->ContentPtr;
         rcaptr->ContentRemaining = rcaptr->EntryPtr->CacheSize;
         rcaptr->ContentLength = 0;
      }

      if (fcptr)
      {
         /* file content buffer (perhaps concurrently with cache buffer) */
         fcptr->CurrentPtr = fcptr->ContentPtr;
         fcptr->ContentRemaining = fcptr->ContentSize;
         fcptr->ContentLength = 0;
      }

      if (!rcaptr->EntryPtr && !fcptr)
      {
         /* not buffering at all, initialize standard output buffer */
         NetWriteInit (rqptr);
      }
   }
   else
   {
      /* subsequent reads */
      tkptr->FileOds.Rab.rab$l_bkt += tkptr->FileOds.Rab.rab$w_usz >> 9;
   }

   if (rcaptr->EntryPtr)
   {
      /* populating a cache buffer, load it progressively */
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "!UL",
                    rcaptr->ContentRemaining);
      tkptr->FileOds.Rab.rab$l_ubf = rcaptr->CurrentPtr;
      if (rcaptr->ContentRemaining > 0xfe00)
         tkptr->FileOds.Rab.rab$w_usz = 0xfe00;
      else
         tkptr->FileOds.Rab.rab$w_usz = rcaptr->ContentRemaining;
   }
   else
   if (fcptr)
   {
      /* only populating a file contents buffer, load it progressively */
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "!UL",
                    fcptr->ContentRemaining);
      tkptr->FileOds.Rab.rab$l_ubf = fcptr->CurrentPtr;
      if (fcptr->ContentRemaining > 0xfe00)
         tkptr->FileOds.Rab.rab$w_usz = 0xfe00;
      else
         tkptr->FileOds.Rab.rab$w_usz = fcptr->ContentRemaining;
   }
   else
   {
      /* standard output buffer, make it a round number of 512 byte blocks */
      BufferSize = OutputBufferSize & 0xfe00;
      tkptr->FileOds.Rab.rab$l_ubf = rqptr->rqOutput.BufferPtr;
      tkptr->FileOds.Rab.rab$w_usz = BufferSize;
   }

   sys$read (&tkptr->FileOds.Rab, &FileNextBlocksAst, &FileNextBlocksAst);
}

/*****************************************************************************/
/*
The read of the next series of Virtual Blocks from the file has completed.  
Post-process and/or queue a network write to the client.  When the network 
write completes it will call the function FileNextBlocks() to queue a read 
of the next series of blocks.
*/ 

FileNextBlocksAst (struct RAB *RabPtr)

{
   int  status, rsz;
   char  *ubf;
   REQUEST_STRUCT  *rqptr;
   FILE_TASK  *tkptr;
   FILE_CONTENT  *fcptr;
   REQUEST_CACHE  *rcaptr;

   /*********/
   /* begin */
   /*********/

   rqptr = RabPtr->rab$l_ctx;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
   {
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
"FileNextBlocksAst() !&F sts:%X!XL stv:%X!XL \
bkt:!UL usz:!UL rsz:!UL EOBIO:!UL",
                 &FileNextBlocksAst, RabPtr->rab$l_sts, RabPtr->rab$l_stv,
                 RabPtr->rab$l_bkt, RabPtr->rab$w_usz, RabPtr->rab$w_rsz,
                 rqptr->FileTaskPtr->EndOfBlockIO);
      if (VMSok (RabPtr->rab$l_sts))
         WatchDataDump (RabPtr->rab$l_ubf, RabPtr->rab$w_rsz);
   }

   rcaptr = &rqptr->rqCache;
   tkptr = rqptr->FileTaskPtr;
   fcptr = rqptr->FileContentPtr;

   if (VMSnok (tkptr->FileOds.Rab.rab$l_sts))
   {
      if (tkptr->FileOds.Rab.rab$l_sts == RMS$_EOF)
      {
         if (rcaptr->EntryPtr)
         {
            /* a cache load becomes valid only after we get to end-of-file */
            rcaptr->EntryPtr->CacheLoadValid = true;
         }
         if (!fcptr)
         {
            /* reset the standard output buffer */
            NetWriteInit (rqptr);
         }
         FileEnd (rqptr);
         return;
      }

      /* legitimate read error */
      rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, tkptr->FileOds.Rab.rab$l_sts, FI_LI);
      FileEnd (rqptr);
      return;
   }

   ubf = (char*)tkptr->FileOds.Rab.rab$l_ubf;
   rsz = tkptr->FileOds.Rab.rab$w_rsz;

   /* if _rsz is less than _usz then it's an effective EOF for block I/O! */
   if (rsz < tkptr->FileOds.Rab.rab$w_usz) tkptr->EndOfBlockIO = true;

   /* when caching or getting file contents we never escape HTML! */
   if (rqptr->FileTaskPtr->EscapeHtml)
   {
      /* queue a network write to the client, AST to FileNextBlocks() */
      rqptr->rqOutput.BufferCount = tkptr->FileOds.Rab.rab$w_rsz;
      FileWriteBufferEscapeHtml (rqptr, &FileNextBlocks);
      return;
   }

   if (rcaptr->EntryPtr)
   {
      /* populating a cache buffer (and possibly also file contents buffer) */
      rcaptr->ContentRemaining -= rsz;
      rcaptr->EntryPtr->ContentLength = (rcaptr->ContentLength += rsz);
      rcaptr->CurrentPtr += rsz;
      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
         WatchThis (rqptr, FI_LI, WATCH_MOD_FILE, "!UL",
                    rcaptr->ContentRemaining);
      if (!rcaptr->ContentRemaining)
         tkptr->EndOfBlockIO = true;
      else
      if (rcaptr->ContentRemaining < 0)
      {
         /*
            This is not supposed to happen!
            Possibly in between getting the file size and opening and
            reading this far the file has been extended or rewritten.
            Shouldn't happen often enough to be a worry!
         */
         rqptr->rqResponse.ErrorTextPtr = MapVmsPath (tkptr->FileName, rqptr);
         rqptr->rqResponse.ErrorOtherTextPtr = tkptr->FileName;
         ErrorVmsStatus (rqptr, RMS$_RTB, FI_LI);
         FileEnd (rqptr);
         return;
      }

      if (!fcptr)
      {
         /* not populating file content buffer, write cache buffer to client */
         NetWrite (rqptr, &FileNextBlocks, ubf, rsz);
         return;
      }

      /* copy from the cache buffer to the file content buffer */
      memcpy (fcptr->CurrentPtr, ubf, rsz);
      fcptr->ContentRemaining -= rsz;
      fcptr->ContentLength += rsz;
      fcptr->CurrentPtr += rsz;
      /* filling the cache/contents buffer, just get more file data */
      SysDclAst (&FileNextBlocks, rqptr);
      return;
   }

   if (fcptr)
   {
      /* populating a file contents buffer, but not a cache buffer */
      fcptr->ContentRemaining -= rsz;
      fcptr->ContentLength += rsz;
      fcptr->CurrentPtr += rsz;
      /* just get more file data */
      SysDclAst (&FileNextBlocks, rqptr);
      return;
   }

   /* write to client (not buffering to cache or to contents) */
   NetWrite (rqptr, &FileNextBlocks, ubf, rsz);
}

/*****************************************************************************/
/*
Send the buffer contents escaping any HTML-forbidden characters.
*/ 

FileWriteBufferEscapeHtml
(
REQUEST_STRUCT *rqptr,
REQUEST_AST AstFunction
)
{
   int  bcnt, ecnt, status;
   char  *bptr, *sptr, *zptr;
   FILE_TASK  *tkptr;
                         
   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_FILE))
      WatchThis (rqptr, FI_LI, WATCH_MOD_FILE,
                 "FileWriteBufferEscapeHtml() !&A !UL",
                 AstFunction, rqptr->rqOutput.BufferCount);

   tkptr = rqptr->FileTaskPtr;

   /* allocate a worst-case escaped HTML buffer (i.e. all "&amp;"s) */
   if (!tkptr->EscapeHtmlPtr)
      tkptr->EscapeHtmlPtr = VmGetHeap (rqptr, OutputBufferSize * 5);

   sptr = tkptr->EscapeHtmlPtr;
   bptr = rqptr->rqOutput.BufferPtr;
   bcnt = rqptr->rqOutput.BufferCount;
   while (bcnt)
   {
      switch (*bptr)
      {
         case '<' :
            memcpy (sptr, "&lt;", 4); sptr += 4; bptr++; bcnt--; continue;
         case '>' :
            memcpy (sptr, "&gt;", 4); sptr += 4; bptr++; bcnt--; continue;
         case '&' :
            memcpy (sptr, "&amp;", 5); sptr += 5; bptr++; bcnt--; continue;
         default :
            *sptr++ = *bptr++; bcnt--;
      }
   }

   NetWrite (rqptr, AstFunction, tkptr->EscapeHtmlPtr,
                                 sptr - tkptr->EscapeHtmlPtr);

   /* reuse standard output buffer */
   NetWriteInit (rqptr);
}

/*****************************************************************************/


